#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function

from aeneas.syncmap.smfbase import SyncMapFormatBase
import aeneas.globalfunctions as gf


class SyncMapFormatGenericSubtitles(SyncMapFormatBase):
    """
    Base class for subtitles-like I/O format handlers.
    """

    TAG = u"SyncMapFormatGenericSubtitles"

    DEFAULT = "subtitles"
    """
    The code for the default variant
    associated with this format.
    """

    def __init__(self, variant=DEFAULT, parameters=None, rconf=None, logger=None):
        super(SyncMapFormatGenericSubtitles, self).__init__(variant=variant, parameters=parameters, rconf=rconf, logger=logger)

        #
        # NOTE since we store functions (parse_..., format_...)
        #      we prefer making these instance members rather than class members
        #
        self.header_string = None
        """
        If not ``None``, the file has the given header.
        """

        self.header_might_not_have_trailing_blank_line = False
        """
        If ``True``, the header line might not be followed by a blank line.
        """

        self.footer_string = None
        """
        If not ``None``, the file has the given footer.
        """

        self.cue_has_identifier = True
        """
        If ``True``, each cue has an identifier in its first line.
        """

        self.cue_has_optional_identifier = False
        """
        If ``True``, each cue might have an identifier in its first line.
        """

        self.time_values_separator = u" --> "
        """
        The separator between time values.
        """

        self.line_break_symbol = u"\n"
        """
        Symbol to place between CC lines.
        """

        self.parse_time_function = gf.time_from_srt
        """
        Function to parse time values.
        """

        self.format_time_function = gf.time_to_srt
        """
        Function to format time values.
        """

    def ignore_block(self, block_lines):
        """
        Return ``True`` if the given block of lines should be ignored.
        """
        return False

    def parse(self, input_text, syncmap):
        def get_block(input_lines, i):
            """
            Get all the non-empty, consecutive lines, starting from index i,
            and then skip all subsequent consecutive empty lines,
            returning the index of the next non-empty line
            (i.e., the line where the next block starts).
            """
            acc = []
            while (i < len(input_lines)) and (input_lines[i] != u""):
                acc.append(input_lines[i])
                i += 1
            while (i < len(input_lines)) and (input_lines[i] == u""):
                i += 1
            return (acc, i)

        def parse_time_string(string):
            """
            Parse begin and end time from the given string.
            """
            split = string.split(self.time_values_separator)
            if len(split) < 2:
                self.log_exc(u"The following timing string is malformed: '%s'" % (string), None, True, ValueError)
            #
            # certain formats might have time lines like:
            # "00:00:20,000 --> 00:00:22,000  X1:40 X2:600 Y1:20 Y2:50"
            # so:
            # begin = "00:00:20,000"
            # end = "00:00:22,000  X1:40 X2:600 Y1:20 Y2:50"
            # so, for end, we need to split over spaces
            # and keep only the first element
            #
            begin, end = split[0:2]
            begin = begin.strip()
            end = ((end.strip()).split(u" "))[0]
            begin = self.parse_time_function(begin)
            end = self.parse_time_function(end)
            return (begin, end)

        input_lines = [l.strip() for l in input_text.splitlines()]
        i = 0
        cue_index = 1

        # skip header, if any
        if self.header_string is not None:
            while i < len(input_lines):
                if input_lines[i].startswith(self.header_string):
                    if self.header_might_not_have_trailing_blank_line:
                        i += 1
                    else:
                        acc, i = get_block(input_lines, i)
                    break
                i += 1

        # skip any blank lines after the header
        while (i < len(input_lines)) and (input_lines[i] == u""):
            i += 1

        # input_lines[i] is not empty
        while i < len(input_lines):
            acc, i = get_block(input_lines, i)

            if len(acc) < 1:
                # no block => break
                break

            if (
                    (self.footer_string is not None) and
                    (acc[0].startswith(self.footer_string))
            ):
                # we reached the footer => break
                break

            if not self.ignore_block(acc):
                # get identifier, if any
                j = 0
                identifier = u"f%06d" % cue_index
                if self.cue_has_identifier:
                    identifier = acc[j]
                    j += 1
                elif self.cue_has_optional_identifier:
                    if self.time_values_separator not in acc[j]:
                        # no time values separator in this line => it is an identifier
                        identifier = acc[j]
                        j += 1

                # get timing string and parse it
                begin, end = parse_time_string(acc[j])
                j += 1

                # get text lines
                lines = (u"\n".join(acc[j:])).split(self.line_break_symbol)

                # append fragment
                self._add_fragment(
                    syncmap=syncmap,
                    identifier=identifier,
                    lines=lines,
                    begin=begin,
                    end=end
                )

                # increase cue index
                cue_index += 1

    def format(self, syncmap):
        msg = []
        if self.header_string is not None:
            msg.append(self.header_string)
            msg.append(u"")
        for i, fragment in enumerate(syncmap.fragments, 1):
            text = fragment.text_fragment
            if self.cue_has_identifier or self.cue_has_optional_identifier:
                msg.append(u"%d" % i)
            msg.append(u"%s%s%s" % (
                self.format_time_function(fragment.begin),
                self.time_values_separator,
                self.format_time_function(fragment.end),
            ))
            lines = self.line_break_symbol.join(text.lines)
            msg.append(lines)
            msg.append(u"")
        if self.footer_string is not None:
            msg.append(self.footer_string)
        else:
            msg.append(u"")
        return u"\n".join(msg)



#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Global common functions.
"""

from __future__ import absolute_import
from __future__ import print_function
import datetime
import io
import math
import os
import re
import shutil
import sys
import tempfile
import uuid

from aeneas.exacttiming import TimeValue
import aeneas.globalconstants as gc


# RUNTIME CONSTANTS

# ANSI codes to color output in terminal
ANSI_END = u"\033[0m"
ANSI_ERROR = u"\033[91m"
ANSI_OK = u"\033[92m"
ANSI_WARNING = u"\033[93m"

# timing regex patterns
HHMMSS_MMM_PATTERN = re.compile(r"([0-9]*):([0-9]*):([0-9]*)\.([0-9]*)")
HHMMSS_MMM_PATTERN_COMMA = re.compile(r"([0-9]*):([0-9]*):([0-9]*),([0-9]*)")

# True if running from a frozen binary (e.g., compiled with pyinstaller)
FROZEN = getattr(sys, "frozen", False)

# True if running under Python 2
PY2 = (sys.version_info[0] == 2)


# COMMON FUNCTIONS

def safe_print(msg):
    """
    Safely print a given Unicode string to stdout,
    possibly replacing characters non-printable
    in the current stdout encoding.

    :param string msg: the message
    """
    try:
        print(msg)
    except UnicodeEncodeError:
        try:
            # NOTE encoding and decoding so that in Python 3 no b"..." is printed
            encoded = msg.encode(sys.stdout.encoding, "replace")
            decoded = encoded.decode(sys.stdout.encoding, "replace")
            print(decoded)
        except (UnicodeDecodeError, UnicodeEncodeError):
            print(u"[ERRO] An unexpected error happened while printing to stdout.")
            print(u"[ERRO] Please check that your file/string encoding matches the shell encoding.")
            print(u"[ERRO] If possible, set your shell encoding to UTF-8 and convert any files with legacy encodings.")


def print_error(msg, color=True):
    """
    Print an error message.

    :param string msg: the message
    :param bool color: if ``True``, print with POSIX color
    """
    if color and is_posix():
        safe_print(u"%s[ERRO] %s%s" % (ANSI_ERROR, msg, ANSI_END))
    else:
        safe_print(u"[ERRO] %s" % (msg))


def print_info(msg, color=True):
    """
    Print an info message.

    :param string msg: the message
    :param bool color: if ``True``, print with POSIX color
    """
    safe_print(u"[INFO] %s" % (msg))


def print_success(msg, color=True):
    """
    Print a success message.

    :param string msg: the message
    :param bool color: if ``True``, print with POSIX color
    """
    if color and is_posix():
        safe_print(u"%s[INFO] %s%s" % (ANSI_OK, msg, ANSI_END))
    else:
        safe_print(u"[INFO] %s" % (msg))


def print_warning(msg, color=True):
    """
    Print a warning message.

    :param string msg: the message
    :param bool color: if ``True``, print with POSIX color
    """
    if color and is_posix():
        safe_print(u"%s[WARN] %s%s" % (ANSI_WARNING, msg, ANSI_END))
    else:
        safe_print(u"[WARN] %s" % (msg))


def uuid_string():
    """
    Return a uuid4 as a Unicode string.

    :rtype: string
    """
    return safe_unicode(str(uuid.uuid4())).lower()


def custom_tmp_dir():
    """
    Return the path of the temporary directory to use.

    On POSIX OSes (Linux and OS X), return the value of
    :data:`~aeneas.globalconstants.TMP_PATH_DEFAULT_POSIX`
    (e.g., ``/tmp/``).

    On non-POSIX OSes, return the value of
    :data:`~aeneas.globalconstants.TMP_PATH_DEFAULT_NONPOSIX`
    (i.e., ``None``), so that ``tempfile``
    will use the directory specified by the
    environment/user ``TMP`` or ``TEMP`` variable.

    :rtype: string
    """
    if is_posix():
        return gc.TMP_PATH_DEFAULT_POSIX
    return gc.TMP_PATH_DEFAULT_NONPOSIX


def tmp_directory(root=None):
    """
    Return the path of a temporary directory created by ``tempfile``.

    :param string root: path to the root temporary directory;
                        if ``None``, the default temporary directory
                        will be used instead
    :rtype: string
    """
    if root is None:
        root = custom_tmp_dir()
    return tempfile.mkdtemp(dir=root)


def tmp_file(suffix=u"", root=None):
    """
    Return a (handler, path) tuple
    for a temporary file with given suffix created by ``tempfile``.

    :param string suffix: the suffix (e.g., the extension) of the file
    :param string root: path to the root temporary directory;
                        if ``None``, the default temporary directory
                        will be used instead
    :rtype: tuple
    """
    if root is None:
        root = custom_tmp_dir()
    return tempfile.mkstemp(suffix=suffix, dir=root)


def file_extension(path):
    """
    Return the file extension.

    Examples: ::

        /foo/bar.baz => baz
        None         => None

    :param string path: the file path
    :rtype: string
    """
    if path is None:
        return None
    ext = os.path.splitext(os.path.basename(path))[1]
    if ext.startswith("."):
        ext = ext[1:]
    return ext


def mimetype_from_path(path):
    """
    Return a mimetype from the file extension.

    :param string path: the file path
    :rtype: string
    """
    extension = file_extension(path)
    if extension is not None:
        extension = extension.lower()
        if extension in gc.MIMETYPE_MAP:
            return gc.MIMETYPE_MAP[extension]
    return None


def file_name_without_extension(path):
    """
    Return the file name without extension.

    Examples: ::

        /foo/bar.baz => bar
        /foo/bar     => bar
        None         => None

    :param string path: the file path
    :rtype: string
    """
    if path is None:
        return None
    return os.path.splitext(os.path.basename(path))[0]


def datetime_string(time_zone=False):
    """
    Return a string representing the current date and time,
    in ``YYYY-MM-DDThh:mm:ss`` or ``YYYY-MM-DDThh:mm:ss+hh:mm`` format

    :param boolean time_zone: if ``True``, add the time zone offset.

    :rtype: string
    """
    time = datetime.datetime.now()
    template = u"%04d-%02d-%02dT%02d:%02d:%02d"
    if time_zone:
        template += u"+00:00"
    return template % (
        time.year,
        time.month,
        time.day,
        time.hour,
        time.minute,
        time.second
    )


def safe_float(string, default=None):
    """
    Safely parse a string into a float.

    On error return the ``default`` value.

    :param string string: string value to be converted
    :param float default: default value to be used in case of failure
    :rtype: float
    """
    value = default
    try:
        value = float(string)
    except TypeError:
        pass
    except ValueError:
        pass
    return value


def safe_int(string, default=None):
    """
    Safely parse a string into an int.

    On error return the ``default`` value.

    :param string string: string value to be converted
    :param int default: default value to be used in case of failure
    :rtype: int
    """
    value = safe_float(string, default)
    if value is not None:
        value = int(value)
    return value


def safe_get(dictionary, key, default_value, can_return_none=True):
    """
    Safely perform a dictionary get,
    returning the default value if the key is not found.

    :param dict dictionary: the dictionary
    :param string key: the key
    :param variant default_value: the default value to be returned
    :param bool can_return_none: if ``True``, the function can return ``None``;
                                 otherwise, return ``default_value`` even if the
                                 dictionary lookup succeeded
    :rtype: variant
    """
    return_value = default_value
    try:
        return_value = dictionary[key]
        if (return_value is None) and (not can_return_none):
            return_value = default_value
    except (KeyError, TypeError):
        # KeyError if key is not present in dictionary
        # TypeError if dictionary is None
        pass
    return return_value


def norm_join(prefix, suffix):
    """
    Join ``prefix`` and ``suffix`` paths
    and return the resulting path, normalized.

    :param string prefix: the prefix path
    :param string suffix: the suffix path
    :rtype: string
    """
    if (prefix is None) and (suffix is None):
        return "."
    if prefix is None:
        return os.path.normpath(suffix)
    if suffix is None:
        return os.path.normpath(prefix)
    return os.path.normpath(os.path.join(prefix, suffix))


def config_txt_to_string(string):
    """
    Convert the contents of a TXT config file
    into the corresponding configuration string ::

        key_1=value_1|key_2=value_2|...|key_n=value_n

    Leading and trailing blank characters will be stripped
    and empty lines (after stripping) will be ignored.

    :param string string: the contents of a TXT config file
    :rtype: string
    """
    if string is None:
        return None
    pairs = [l.strip() for l in string.splitlines() if len(l.strip()) > 0]
    return gc.CONFIG_STRING_SEPARATOR_SYMBOL.join(pairs)


def config_string_to_dict(string, result=None):
    """
    Convert a given configuration string ::

        key_1=value_1|key_2=value_2|...|key_n=value_n

    into the corresponding dictionary ::

        dictionary[key_1] = value_1
        dictionary[key_2] = value_2
        ...
        dictionary[key_n] = value_n

    :param string string: the configuration string
    :rtype: dict
    """
    if string is None:
        return {}
    pairs = string.split(gc.CONFIG_STRING_SEPARATOR_SYMBOL)
    return pairs_to_dict(pairs, result)


def config_xml_to_dict(contents, result, parse_job=True):
    """
    Convert the contents of a XML config file
    into the corresponding dictionary ::

        dictionary[key_1] = value_1
        dictionary[key_2] = value_2
        ...
        dictionary[key_n] = value_n

    :param bytes contents: the XML configuration contents
    :param bool parse_job: if ``True``, parse the job properties;
                           if ``False``, parse the tasks properties
    :rtype: dict (``parse_job=True``) or list of dict (``parse_job=False``)
    """
    from lxml import etree
    try:
        root = etree.fromstring(contents)
        pairs = []
        if parse_job:
            # parse job
            for elem in root:
                if (elem.tag != gc.CONFIG_XML_TASKS_TAG) and (elem.text is not None):
                    pairs.append(u"%s%s%s" % (
                        safe_unicode(elem.tag),
                        gc.CONFIG_STRING_ASSIGNMENT_SYMBOL,
                        safe_unicode(elem.text.strip())
                    ))
            return pairs_to_dict(pairs)
        else:
            # parse tasks
            output_list = []
            for task in root.find(gc.CONFIG_XML_TASKS_TAG):
                if task.tag == gc.CONFIG_XML_TASK_TAG:
                    pairs = []
                    for elem in task:
                        if elem.text is not None:
                            pairs.append(u"%s%s%s" % (
                                safe_unicode(elem.tag),
                                gc.CONFIG_STRING_ASSIGNMENT_SYMBOL,
                                safe_unicode(elem.text.strip())
                            ))
                    output_list.append(pairs_to_dict(pairs))
            return output_list
    except:
        if result is not None:
            result.passed = False
            result.add_error("An error occurred while parsing XML file")
        if parse_job:
            return {}
        else:
            return []


def config_dict_to_string(dictionary):
    """
    Convert a given config dictionary ::

        dictionary[key_1] = value_1
        dictionary[key_2] = value_2
        ...
        dictionary[key_n] = value_n

    into the corresponding string ::

        key_1=value_1|key_2=value_2|...|key_n=value_n

    :param dict dictionary: the config dictionary
    :rtype: string
    """
    parameters = []
    for key in dictionary:
        parameters.append(u"%s%s%s" % (
            key,
            gc.CONFIG_STRING_ASSIGNMENT_SYMBOL,
            dictionary[key]
        ))
    return gc.CONFIG_STRING_SEPARATOR_SYMBOL.join(parameters)


def pairs_to_dict(pairs, result=None):
    """
    Convert a given list of ``key=value`` strings ::

        ["key_1=value_1", "key_2=value_2", ..., "key_n=value_n"]

    into the corresponding dictionary ::

        dictionary[key_1] = value_1
        dictionary[key_2] = value_2
        ...
        dictionary[key_n] = value_n

    :param list pairs: the list of key=value strings
    :rtype: dict
    """
    dictionary = {}
    for pair in pairs:
        if len(pair) > 0:
            tokens = pair.split(gc.CONFIG_STRING_ASSIGNMENT_SYMBOL)
            if ((len(tokens) == 2) and
                    (len(tokens[0])) > 0 and
                    (len(tokens[1]) > 0)):
                dictionary[tokens[0]] = tokens[1]
            elif result is not None:
                result.add_warning("Invalid key=value string: '%s'" % pair)
    return dictionary


def copytree(source_directory, destination_directory, ignore=None):
    """
    Recursively copy the contents of a source directory
    into a destination directory.
    Both directories must exist.

    This function does not copy the root directory ``source_directory``
    into ``destination_directory``.

    Since ``shutil.copytree(src, dst)`` requires ``dst`` not to exist,
    we cannot use for our purposes.

    Code adapted from http://stackoverflow.com/a/12686557

    :param string source_directory: the source directory, already existing
    :param string destination_directory: the destination directory, already existing
    """
    if os.path.isdir(source_directory):
        if not os.path.isdir(destination_directory):
            os.makedirs(destination_directory)
        files = os.listdir(source_directory)
        if ignore is not None:
            ignored = ignore(source_directory, files)
        else:
            ignored = set()
        for f in files:
            if f not in ignored:
                copytree(
                    os.path.join(source_directory, f),
                    os.path.join(destination_directory, f),
                    ignore
                )
    else:
        shutil.copyfile(source_directory, destination_directory)


def ensure_parent_directory(path, ensure_parent=True):
    """
    Ensures the parent directory exists.

    :param string path: the path of the file
    :param bool ensure_parent: if ``True``, ensure the parent directory of ``path`` exists;
                               if ``False``, ensure ``path`` exists
    :raises: OSError: if the path cannot be created
    """
    parent_directory = os.path.abspath(path)
    if ensure_parent:
        parent_directory = os.path.dirname(parent_directory)
    if not os.path.exists(parent_directory):
        try:
            os.makedirs(parent_directory)
        except (IOError, OSError):
            raise OSError(u"Directory '%s' cannot be created" % parent_directory)


def time_from_ttml(string):
    """
    Parse the given ``SS.mmms`` string
    (TTML values have an "s" suffix, e.g. ``1.234s``)
    and return a time value.

    :param string string: the string to be parsed
    :rtype: :class:`~aeneas.exacttiming.TimeValue`
    """
    if (string is None) or (len(string) < 2):
        return 0
    # strips "s" at the end
    string = string[:-1]
    return time_from_ssmmm(string)


def time_to_ttml(time_value):
    """
    Format the given time value into a ``SS.mmms`` string
    (TTML values have an "s" suffix, e.g. ``1.234s``).

    Examples: ::

        12        => 12.000s
        12.345    => 12.345s
        12.345432 => 12.345s
        12.345678 => 12.346s

    :param float time_value: a time value, in seconds
    :rtype: string
    """
    if time_value is None:
        time_value = 0
    return "%ss" % time_to_ssmmm(time_value)


def time_from_ssmmm(string):
    """
    Parse the given ``SS.mmm`` string and return a time value.

    :param string string: the string to be parsed
    :rtype: :class:`~aeneas.exacttiming.TimeValue`
    """
    if (string is None) or (len(string) < 1):
        return TimeValue("0.000")
    return TimeValue(string)


def time_to_ssmmm(time_value):
    """
    Format the given time value into a ``SS.mmm`` string.

    Examples: ::

        12        => 12.000
        12.345    => 12.345
        12.345432 => 12.345
        12.345678 => 12.346

    :param float time_value: a time value, in seconds
    :rtype: string
    """
    if time_value is None:
        time_value = 0
    return "%.3f" % (time_value)


def time_from_hhmmssmmm(string, decimal_separator="."):
    """
    Parse the given ``HH:MM:SS.mmm`` string and return a time value.

    :param string string: the string to be parsed
    :param string decimal_separator: the decimal separator to be used
    :rtype: :class:`~aeneas.exacttiming.TimeValue`
    """
    if decimal_separator == ",":
        pattern = HHMMSS_MMM_PATTERN_COMMA
    else:
        pattern = HHMMSS_MMM_PATTERN
    v_length = TimeValue("0.000")
    try:
        match = pattern.search(string)
        if match is not None:
            v_h = int(match.group(1))
            v_m = int(match.group(2))
            v_s = int(match.group(3))
            v_f = TimeValue("0." + match.group(4))
            v_length = v_h * 3600 + v_m * 60 + v_s + v_f
    except:
        pass
    return v_length


def time_to_hhmmssmmm(time_value, decimal_separator="."):
    """
    Format the given time value into a ``HH:MM:SS.mmm`` string.

    Examples: ::

        12        => 00:00:12.000
        12.345    => 00:00:12.345
        12.345432 => 00:00:12.345
        12.345678 => 00:00:12.346
        83        => 00:01:23.000
        83.456    => 00:01:23.456
        83.456789 => 00:01:23.456
        3600      => 01:00:00.000
        3612.345  => 01:00:12.345

    :param float time_value: a time value, in seconds
    :param string decimal_separator: the decimal separator, default ``.``
    :rtype: string
    """
    if time_value is None:
        time_value = 0
    tmp = time_value
    hours = int(math.floor(tmp / 3600))
    tmp -= (hours * 3600)
    minutes = int(math.floor(tmp / 60))
    tmp -= minutes * 60
    seconds = int(math.floor(tmp))
    tmp -= seconds
    milliseconds = int(math.floor(tmp * 1000))
    return "%02d:%02d:%02d%s%03d" % (
        hours,
        minutes,
        seconds,
        decimal_separator,
        milliseconds
    )


def time_from_srt(string):
    """
    Parse the given ``HH:MM:SS,mmm`` string and return a time value.

    :param string string: the string to be parsed
    :rtype: :class:`~aeneas.exacttiming.TimeValue`
    """
    return time_from_hhmmssmmm(string, decimal_separator=",")


def time_to_srt(time_value):
    """
    Format the given time value into a ``HH:MM:SS,mmm`` string,
    as used in the SRT format.

    Examples: ::

        12        => 00:00:12,000
        12.345    => 00:00:12,345
        12.345432 => 00:00:12,345
        12.345678 => 00:00:12,346
        83        => 00:01:23,000
        83.456    => 00:01:23,456
        83.456789 => 00:01:23,456
        3600      => 01:00:00,000
        3612.345  => 01:00:12,345

    :param float time_value: a time value, in seconds
    :rtype: string
    """
    return time_to_hhmmssmmm(time_value, decimal_separator=",")


def split_url(url):
    """
    Split the given URL ``base#anchor`` into ``(base, anchor)``,
    or ``(base, None)`` if no anchor is present.

    In case there are two or more ``#`` characters,
    return only the first two tokens: ``a#b#c => (a, b)``.

    :param string url: the url
    :rtype: list of str
    """
    if url is None:
        return (None, None)
    array = url.split("#")
    if len(array) == 1:
        array.append(None)
    return tuple(array[0:2])


def is_posix():
    """
    Return ``True`` if running on a POSIX OS.

    :rtype: bool
    """
    # from https://docs.python.org/2/library/os.html#os.name
    # the registered values of os.name are:
    # "posix", "nt", "os2", "ce", "java", "riscos"
    return os.name == "posix"


def is_linux():
    """
    Return ``True`` if running on Linux.

    :rtype: bool
    """
    return (is_posix()) and (os.uname()[0] == "Linux")


def is_osx():
    """
    Return ``True`` if running on Mac OS X (Darwin).

    :rtype: bool
    """
    return (is_posix()) and (os.uname()[0] == "Darwin")


def is_windows():
    """
    Return ``True`` if running on Windows.

    :rtype: bool
    """
    return os.name == "nt"


def is_py2_narrow_build():
    """
    Return ``True`` if running on a Python 2 narrow build.

    :rtype: bool
    """
    return (PY2) and (sys.maxunicode == 65535)


def fix_slash(path):
    """
    On non-POSIX OSes, change the slashes in ``path``
    for loading in the browser.

    Example: ::

        c:\\abc\\def => c:/abc/def

    :param string path: the path
    :rtype: string
    """
    if not is_posix():
        path = path.replace("\\", "/")
    return path


def can_run_c_extension(name=None):
    """
    Determine whether the given Python C extension loads correctly.

    If ``name`` is ``None``, tests all Python C extensions,
    and return ``True`` if and only if all load correctly.

    :param string name: the name of the Python C extension to test
    :rtype: bool
    """
    def can_run_cdtw():
        """ Python C extension for computing DTW """
        try:
            import aeneas.cdtw.cdtw
            return True
        except ImportError:
            return False

    def can_run_cmfcc():
        """ Python C extension for computing MFCC """
        try:
            import aeneas.cmfcc.cmfcc
            return True
        except ImportError:
            return False

    def can_run_cew():
        """ Python C extension for synthesizing with eSpeak """
        try:
            import aeneas.cew.cew
            return True
        except ImportError:
            return False

    def can_run_cfw():
        """ Python C extension for synthesizing with Festival """
        try:
            import aeneas.cfw.cfw
            return True
        except ImportError:
            return False

    if name == "cdtw":
        return can_run_cdtw()
    elif name == "cmfcc":
        return can_run_cmfcc()
    elif name == "cew":
        return can_run_cew()
    elif name == "cfw":
        return can_run_cfw()
    else:
        # NOTE cfw is still experimental!
        return can_run_cdtw() and can_run_cmfcc() and can_run_cew()


def run_c_extension_with_fallback(
        log_function,
        extension,
        c_function,
        py_function,
        args,
        rconf
):
    """
    Run a function calling a C extension, falling back
    to a pure Python function if the former does not succeed.

    :param function log_function: a logger function
    :param string extension: the name of the extension
    :param function c_function: the (Python) function calling the C extension
    :param function py_function: the (Python) function providing the fallback
    :param rconf: the runtime configuration
    :type  rconf: :class:`aeneas.runtimeconfiguration.RuntimeConfiguration`
    :rtype: depends on the extension being called
    :raises: RuntimeError: if both the C extension and
                           the pure Python code did not succeed.

    .. versionadded:: 1.4.0
    """
    computed = False
    if not rconf[u"c_extensions"]:
        log_function(u"C extensions disabled")
    elif extension not in rconf:
        log_function([u"C extension '%s' not recognized", extension])
    elif not rconf[extension]:
        log_function([u"C extension '%s' disabled", extension])
    else:
        log_function([u"C extension '%s' enabled", extension])
        if c_function is None:
            log_function(u"C function is None")
        elif can_run_c_extension(extension):
            log_function([u"C extension '%s' enabled and it can be loaded", extension])
            computed, result = c_function(*args)
        else:
            log_function([u"C extension '%s' enabled but it cannot be loaded", extension])
    if not computed:
        if py_function is None:
            log_function(u"Python function is None")
        else:
            log_function(u"Running the pure Python code")
            computed, result = py_function(*args)
    if not computed:
        raise RuntimeError(u"Both the C extension and the pure Python code failed. (Wrong arguments? Input too big?)")
    return result


def file_can_be_read(path):
    """
    Return ``True`` if the file at the given ``path`` can be read.

    :param string path: the file path
    :rtype: bool

    .. versionadded:: 1.4.0
    """
    if path is None:
        return False
    try:
        with io.open(path, "rb") as test_file:
            pass
        return True
    except (IOError, OSError):
        pass
    return False


def file_can_be_written(path):
    """
    Return ``True`` if a file can be written at the given ``path``.

    :param string path: the file path
    :rtype: bool

    .. warning:: This function will attempt to open the given ``path``
                 in write mode, possibly destroying the file previously existing there.

    .. versionadded:: 1.4.0
    """
    if path is None:
        return False
    try:
        with io.open(path, "wb") as test_file:
            pass
        delete_file(None, path)
        return True
    except (IOError, OSError):
        pass
    return False


def directory_exists(path):
    """
    Return ``True`` if the given ``path`` string
    points to an existing directory.

    :param string path: the file path
    :rtype: bool
    """
    if (path is None) or (not os.path.isdir(path)):
        return False
    return True


def file_exists(path):
    """
    Return ``True`` if the given ``path`` string
    points to an existing file.

    :param string path: the file path
    :rtype: bool
    """
    if (path is None) or (not os.path.isfile(path)):
        return False
    return True


def file_size(path):
    """
    Return the size, in bytes, of the file at the given ``path``.
    Return ``-1`` if the file does not exist or cannot be read.

    :param string path: the file path
    :rtype: int
    """
    try:
        return os.path.getsize(path)
    except OSError:
        return -1


def delete_directory(path):
    """
    Safely delete a directory.

    :param string path: the file path
    """
    if path is not None:
        try:
            shutil.rmtree(path)
        except:
            pass


def close_file_handler(handler):
    """
    Safely close the given file handler.

    :param object handler: the file handler (as returned by tempfile)
    """
    if handler is not None:
        try:
            os.close(handler)
        except:
            pass


def delete_file(handler, path):
    """
    Safely delete file.

    :param object handler: the file handler (as returned by tempfile)
    :param string path: the file path
    """
    close_file_handler(handler)
    if path is not None:
        try:
            os.remove(path)
        except:
            pass


def relative_path(path, from_file):
    """
    Return the relative path of a file or directory, specified
    as ``path`` relative to (the parent directory of) ``from_file``.

    This method is intented to be called with ``__file__``
    as second argument.

    The returned path is relative to the current working directory.

    If ``path`` is ``None``, return ``None``.

    Example: ::

        path="res/foo.bar"
        from_file="/root/abc/def/ghi.py"
        cwd="/root"
        => "abc/def/res/foo.bar"

    :param string path: the file path
    :param string from_file: the reference file
    :rtype: string
    """
    if path is None:
        return None
    abs_path_target = absolute_path(path, from_file)
    abs_path_cwd = os.getcwd()
    if is_windows():
        # NOTE on Windows, if the two paths are on different drives,
        #      the notion of relative path is not defined:
        #      return the absolute path of the target instead.
        t_drive, t_tail = os.path.splitdrive(abs_path_target)
        c_drive, c_tail = os.path.splitdrive(abs_path_cwd)
        if t_drive != c_drive:
            return abs_path_target
    return os.path.relpath(abs_path_target, start=abs_path_cwd)


def absolute_path(path, from_file):
    """
    Return the absolute path of a file or directory, specified
    as ``path`` relative to (the parent directory of) ``from_file``.

    This method is intented to be called with ``__file__``
    as second argument.

    If ``path`` is ``None``, return ``None``.

    Example: ::

        path="res/foo.bar"
        from_file="/abc/def/ghi.py"
        => "/abc/def/res/foo.bar"

    :param string path: the file path
    :param string from_file: the reference file
    :rtype: string
    """
    if path is None:
        return None
    current_directory = os.path.dirname(from_file)
    target = os.path.join(current_directory, path)
    return os.path.abspath(target)


def read_file_bytes(input_file_path):
    """
    Read the file at the given file path
    and return its contents as a byte string,
    or ``None`` if an error occurred.

    :param string input_file_path: the file path
    :rtype: bytes
    """
    contents = None
    try:
        with io.open(input_file_path, "rb") as input_file:
            contents = input_file.read()
    except:
        pass
    return contents


def human_readable_number(number, suffix=""):
    """
    Format the given number into a human-readable string.

    Code adapted from http://stackoverflow.com/a/1094933

    :param variant number: the number (int or float)
    :param string suffix: the unit of the number
    :rtype: string
    """
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(number) < 1024.0:
            return "%3.1f%s%s" % (number, unit, suffix)
        number /= 1024.0
    return "%.1f%s%s" % (number, "Y", suffix)


def is_unicode(string):
    """
    Return ``True`` if the given string is a sequence of Unicode code points.

    :param variant string: the string to test
    :rtype: bool
    """
    if PY2:
        return isinstance(string, unicode)
    return isinstance(string, str)


def is_bytes(string):
    """
    Return ``True`` if the given string is a sequence of bytes.

    :param variant string: the string to test
    :rtype: bool
    """
    if PY2:
        return isinstance(string, str)
    return isinstance(string, bytes)


def is_utf8_encoded(bstring):
    """
    Return ``True`` if the given byte string can be decoded
    into a Unicode string using the UTF-8 decoder.

    :param bytes bstring: the string to test
    :rtype: bool
    """
    try:
        bstring.decode("utf-8")
        return True
    except UnicodeDecodeError:
        pass
    return False


def safe_str(string):
    """
    Safely return the given Unicode string
    from a ``__str__`` function: as a byte string
    in Python 2, or as a Unicode string in Python 3.

    :param string string: the string to return
    :rtype: bytes or string
    """
    if string is None:
        return None
    if PY2:
        return string.encode("utf-8")
    return string


def safe_unichr(codepoint):
    """
    Safely return a Unicode string of length one,
    containing the Unicode character with given codepoint.

    :param int codepoint: the codepoint
    :rtype: string
    """
    if is_py2_narrow_build():
        return ("\\U%08x" % codepoint).decode("unicode-escape")
    elif PY2:
        return unichr(codepoint)
    return chr(codepoint)


def safe_unicode(string):
    """
    Safely convert the given string to a Unicode string.

    :param variant string: the byte string or Unicode string to convert
    :rtype: string
    """
    if string is None:
        return None
    if is_bytes(string):
        return string.decode("utf-8")
    return string


def safe_bytes(string):
    """
    Safely convert the given string to a bytes string.

    :param variant string: the byte string or Unicode string to convert
    :rtype: bytes
    """
    if string is None:
        return None
    if is_unicode(string):
        return string.encode("utf-8")
    return string


def safe_unicode_stdin(string):
    """
    Safely convert the given string to a Unicode string,
    decoding using ``sys.stdin.encoding`` if needed.

    If running from a frozen binary, ``utf-8`` encoding is assumed.

    :param variant string: the byte string or Unicode string to convert
    :rtype: string
    """
    if string is None:
        return None
    if is_bytes(string):
        if FROZEN:
            return string.decode("utf-8")
        try:
            return string.decode(sys.stdin.encoding)
        except UnicodeDecodeError:
            return string.decode(sys.stdin.encoding, "replace")
        except:
            return string.decode("utf-8")
    return string


def object_to_unicode(obj):
    """
    Return a sequence of Unicode code points from the given object.

    :param object obj: the object
    :rtype: string
    """
    if PY2:
        return unicode(obj)
    return str(obj)


def object_to_bytes(obj):
    """
    Return a sequence of bytes from the given object.

    :param object obj: the object
    :rtype: bytes
    """
    if PY2:
        return str(obj)
    return bytes(obj, encoding="utf-8")


def bundle_directory():
    """
    Return the absolute path of the bundle directory
    if running from a frozen binary; otherwise return ``None``.

    :rtype: string
    """
    if FROZEN:
        return sys._MEIPASS
    return None



#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function

from aeneas.syncmap.smfaudacity import SyncMapFormatAudacity
from aeneas.syncmap.smfcsv import SyncMapFormatCSV
from aeneas.syncmap.smfeaf import SyncMapFormatEAF
from aeneas.syncmap.smfjson import SyncMapFormatJSON
from aeneas.syncmap.smfrbse import SyncMapFormatRBSE
from aeneas.syncmap.smfsmil import SyncMapFormatSMIL
from aeneas.syncmap.smfsrt import SyncMapFormatSRT
from aeneas.syncmap.smfssv import SyncMapFormatSSV
from aeneas.syncmap.smfsub import SyncMapFormatSUB
from aeneas.syncmap.smftextgrid import SyncMapFormatTextGrid
from aeneas.syncmap.smftsv import SyncMapFormatTSV
from aeneas.syncmap.smfttml import SyncMapFormatTTML
from aeneas.syncmap.smftxt import SyncMapFormatTXT
from aeneas.syncmap.smfvtt import SyncMapFormatVTT
from aeneas.syncmap.smfxml import SyncMapFormatXML
from aeneas.syncmap.smfxmllegacy import SyncMapFormatXMLLegacy


class SyncMapFormat(object):
    """
    Enumeration of the supported output formats
    for a synchronization map.
    """

    AUD = SyncMapFormatAudacity.DEFAULT
    """
    Alias for ``AUDM``.
    """

    AUDH = SyncMapFormatAudacity.HUMAN
    """
    Tab-separated plain text,
    with human-readable time values
    and fragment text::

        00:00:00.000   00:00:01.234   Text of fragment 1
        00:00:01.234   00:00:05.678   Text of fragment 2
        00:00:05.678   00:00:07.890   Text of fragment 3

    * Multiple levels: no
    * Multiple lines: no

    See also http://manual.audacityteam.org/man/label_tracks.html#export

    .. versionadded:: 1.5.0
    """

    AUDM = SyncMapFormatAudacity.MACHINE
    """
    Tab-separated plain text,
    with machine-readable time values
    and fragment text,
    compatible with ``Audacity``::

        0.000   1.234   Text fragment 1
        1.234   5.678   Text fragment 2
        5.678   7.890   Text fragment 3

    * Multiple levels: no
    * Multiple lines: no

    See also http://manual.audacityteam.org/man/label_tracks.html#export

    .. versionadded:: 1.5.0
    """

    CSV = SyncMapFormatCSV.DEFAULT
    """
    Alias for ``CSVM``.
    """

    CSVH = SyncMapFormatCSV.HUMAN
    """
    Comma-separated values (CSV),
    with human-readable time values::

        f001,00:00:00.000,00:00:01.234,"First fragment text"
        f002,00:00:01.234,00:00:05.678,"Second fragment text"
        f003,00:00:05.678,00:00:07.890,"Third fragment text"

    * Multiple levels: no
    * Multiple lines: no

    Please note that the text is assumed to be contained
    in double quotes ("..."),
    which are stripped when reading from file,
    and added back when writing to file.

    .. versionadded:: 1.0.4
    """

    CSVM = SyncMapFormatCSV.MACHINE
    """
    Comma-separated values (CSV),
    with machine-readable time values::

        f001,0.000,1.234,"First fragment text"
        f002,1.234,5.678,"Second fragment text"
        f003,5.678,7.890,"Third fragment text"

    * Multiple levels: no
    * Multiple lines: no

    Please note that the text is assumed to be contained
    in double quotes ("..."),
    which are stripped when reading from file,
    and added back when writing to file.

    .. versionadded:: 1.2.0
    """

    DFXP = SyncMapFormatTTML.DFXP
    """
    Alias for ``TTML``.

    .. versionadded:: 1.4.1
    """

    EAF = SyncMapFormatEAF.DEFAULT
    """
    ELAN EAF::

        <?xml version="1.0" encoding="UTF-8"?>
        <ANNOTATION_DOCUMENT AUTHOR="aeneas" DATE="2016-01-01T00:00:00+00:00" FORMAT="2.8" VERSION="2.8" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://www.mpi.nl/tools/elan/EAFv2.8.xsd">
            <HEADER MEDIA_FILE="" TIME_UNITS="milliseconds" />
            <TIME_ORDER>
                <TIME_SLOT TIME_SLOT_ID="ts001b" TIME_VALUE="0"/>
                <TIME_SLOT TIME_SLOT_ID="ts001e" TIME_VALUE="1234"/>
                <TIME_SLOT TIME_SLOT_ID="ts002b" TIME_VALUE="1234"/>
                <TIME_SLOT TIME_SLOT_ID="ts002e" TIME_VALUE="5678"/>
                <TIME_SLOT TIME_SLOT_ID="ts003b" TIME_VALUE="5678"/>
                <TIME_SLOT TIME_SLOT_ID="ts003e" TIME_VALUE="7890"/>
            </TIME_ORDER>
            <TIER LINGUISTIC_TYPE_REF="utterance" TIER_ID="tier1">
                <ANNOTATION>
                    <ALIGNABLE_ANNOTATION ANNOTATION_ID="f001" TIME_SLOT_REF1="ts001b" TIME_SLOT_REF2="ts001e">
                        <ANNOTATION_VALUE>First fragment text</ANNOTATION_VALUE>
                    </ALIGNABLE_ANNOTATION>
                </ANNOTATION>
                <ANNOTATION>
                    <ALIGNABLE_ANNOTATION ANNOTATION_ID="f002" TIME_SLOT_REF1="ts002b" TIME_SLOT_REF2="ts002e">
                        <ANNOTATION_VALUE>First fragment text</ANNOTATION_VALUE>
                    </ALIGNABLE_ANNOTATION>
                </ANNOTATION>
                <ANNOTATION>
                    <ALIGNABLE_ANNOTATION ANNOTATION_ID="f003" TIME_SLOT_REF1="ts003b" TIME_SLOT_REF2="ts003e">
                        <ANNOTATION_VALUE>First fragment text</ANNOTATION_VALUE>
                    </ALIGNABLE_ANNOTATION>
                </ANNOTATION>
            </TIER>
            <LINGUISTIC_TYPE LINGUISTIC_TYPE_ID="utterance" TIME_ALIGNABLE="true"/>
        </ANNOTATION_DOCUMENT>

    * Multiple levels: no
    * Multiple lines: no

    See also https://tla.mpi.nl/tla-news/documentation-of-eaf-elan-annotation-format/

    .. versionadded:: 1.5.0
    """

    JSON = SyncMapFormatJSON.DEFAULT
    """
    JSON::

        {
         "fragments": [
          {
           "id": "f001",
           "language": "en",
           "begin": 0.000,
           "end": 1.234,
           "children": [],
           "lines": [
            "First fragment text"
           ]
          },
          {
           "id": "f002",
           "language": "en",
           "begin": 1.234,
           "end": 5.678,
           "children": [],
           "lines": [
            "Second fragment text",
            "Second line of second fragment"
           ]
          },
          {
           "id": "f003",
           "language": "en",
           "begin": 5.678,
           "end": 7.890,
           "children": [],
           "lines": [
            "Third fragment text",
            "Second line of third fragment"
           ]
          }
         ]
        }

    * Multiple levels: yes (output only)
    * Multiple lines: yes

    .. versionadded:: 1.2.0
    """

    RBSE = SyncMapFormatRBSE.DEFAULT
    """
    JSON compatible with ``rb_smil_emulator.js``::

        {
         "smil_ids": [
          "f001",
          "f002",
          "f003",
         ],
         "smil_data": [
          { "id": "f001", "begin": 0.000, "end": 1.234 },
          { "id": "f002", "begin": 1.234, "end": 5.678 },
          { "id": "f003", "begin": 5.678, "end": 7.890 }
         ]
        }

    * Multiple levels: no
    * Multiple lines: no

    See also https://github.com/pettarin/rb_smil_emulator

    Deprecated, it will be removed in v2.0.0.

    .. deprecated:: 1.5.0

    .. versionadded:: 1.2.0
    """

    SBV = SyncMapFormatSUB.SBV
    """
    SubViewer (SBV/SUB) caption/subtitle format,
    with multiple lines per fragment are separated by a newline character::

        [SUBTITLE]
        00:00:00.000,00:00:01.234
        First fragment text

        00:00:01.234,00:00:05.678
        Second fragment text
        Second line of second fragment

        00:00:05.678,00:00:07.890
        Third fragment text
        Second line of third fragment

    * Multiple levels: no
    * Multiple lines: yes

    See also https://wiki.videolan.org/SubViewer/

    Note that the ``[INFORMATION]`` header is ignored when reading,
    and it is not produced when writing.
    Moreover, extensions
    (i.e., ``[COLF]``, ``[SIZE]``, ``[FONT]``)
    are not supported.
    """

    SMIL = SyncMapFormatSMIL.DEFAULT
    """
    Alias for ``SMILH``.
    """

    SMILH = SyncMapFormatSMIL.HUMAN
    """
    SMIL (as in the EPUB 3 Media Overlay specification),
    with human-readable time values::

        <smil xmlns="http://www.w3.org/ns/SMIL" xmlns:epub="http://www.idpf.org/2007/ops" version="3.0">
         <body>
          <seq id="seq000001" epub:textref="p001.xhtml">
           <par id="par000001">
            <text src="p001.xhtml#f001"/>
            <audio clipBegin="00:00:00.000" clipEnd="00:00:01.234" src="../Audio/p001.mp3"/>
           </par>
           <par id="par000002">
            <text src="p001.xhtml#f002"/>
            <audio clipBegin="00:00:01.234" clipEnd="00:00:05.678" src="../Audio/p001.mp3"/>
           </par>
           <par id="par000003">
            <text src="p001.xhtml#f003"/>
            <audio clipBegin="00:00:05.678" clipEnd="00:00:07.890" src="../Audio/p001.mp3"/>
           </par>
          </seq>
         </body>
        </smil>

    * Multiple levels: yes (output only)
    * Multiple lines: no

    See also http://www.idpf.org/epub3/latest/mediaoverlays

    .. versionadded:: 1.2.0
    """

    SMILM = SyncMapFormatSMIL.MACHINE
    """
    SMIL (as in the EPUB 3 Media Overlay specification),
    with machine-readable time values::

        <smil xmlns="http://www.w3.org/ns/SMIL" xmlns:epub="http://www.idpf.org/2007/ops" version="3.0">
         <body>
          <seq id="seq000001" epub:textref="p001.xhtml">
           <par id="par000001">
            <text src="p001.xhtml#f001"/>
            <audio clipBegin="0.000" clipEnd="1.234" src="../Audio/p001.mp3"/>
           </par>
           <par id="par000002">
            <text src="p001.xhtml#f002"/>
            <audio clipBegin="1.234" clipEnd="5.678" src="../Audio/p001.mp3"/>
           </par>
           <par id="par000003">
            <text src="p001.xhtml#f003"/>
            <audio clipBegin="5.678" clipEnd="7.890" src="../Audio/p001.mp3"/>
           </par>
          </seq>
         </body>
        </smil>

    * Multiple levels: yes (output only)
    * Multiple lines: no

    See also http://www.idpf.org/epub3/latest/mediaoverlays

    .. versionadded:: 1.2.0
    """

    SRT = SyncMapFormatSRT.DEFAULT
    """
    SubRip (SRT) caption/subtitle format
    (it might have multiple lines per fragment)::

        1
        00:00:00,000 --> 00:00:01,234
        First fragment text

        2
        00:00:01,234 --> 00:00:05,678
        Second fragment text
        Second line of second fragment

        3
        00:00:05,678 --> 00:00:07,890
        Third fragment text
        Second line of third fragment

    * Multiple levels: no
    * Multiple lines: yes

    See also https://wiki.videolan.org/SubRip/

    Note that extensions
    (i.e., ``<b>``, ``<s>``, ``<u>``, ``<i>``, ``<font>``)
    are not supported.
    """

    SSV = SyncMapFormatSSV.DEFAULT
    """
    Alias for ``SSVM``.

    .. versionadded:: 1.0.4
    """

    SSVH = SyncMapFormatSSV.HUMAN
    """
    Space-separated plain text,
    with human-readable time values::

        00:00:00.000 00:00:01.234 f001 "First fragment text"
        00:00:01.234 00:00:05.678 f002 "Second fragment text"
        00:00:05.678 00:00:07.890 f003 "Third fragment text"

    * Multiple levels: no
    * Multiple lines: no

    Please note that the text is assumed to be contained
    in double quotes ("..."),
    which are stripped when reading from file,
    and added back when writing to file.

    .. versionadded:: 1.0.4
    """

    SSVM = SyncMapFormatSSV.MACHINE
    """
    Space-separated plain text,
    with machine-readable time values::

        0.000 1.234 f001 "First fragment text"
        1.234 5.678 f002 "Second fragment text"
        5.678 7.890 f003 "Third fragment text"

    * Multiple levels: no
    * Multiple lines: no

    Please note that the text is assumed to be contained
    in double quotes ("..."),
    which are stripped when reading from file,
    and added back when writing to file.

    .. versionadded:: 1.2.0
    """

    SUB = SyncMapFormatSUB.SUB
    """
    SubViewer (SBV/SUB) caption/subtitle format,
    with multiple lines per fragment are separated by [br]::

        [SUBTITLE]
        00:00:00.000,00:00:01.234
        First fragment text

        00:00:01.234,00:00:05.678
        Second fragment text[br]Second line of second fragment

        00:00:05.678,00:00:07.890
        Third fragment text[br]Second line of third fragment

    * Multiple levels: no
    * Multiple lines: yes

    See also https://wiki.videolan.org/SubViewer/

    Note that the ``[INFORMATION]`` header is ignored when reading,
    and it is not produced when writing.
    Moreover, extensions
    (i.e., ``[COLF]``, ``[SIZE]``, ``[FONT]``)
    are not supported.

    .. versionadded:: 1.4.1
    """

    TAB = SyncMapFormatTSV.TAB
    """
    Deprecated, it will be removed in v2.0.0.
    Use ``TSV`` instead.

    .. deprecated:: 1.0.3
    """

    TEXTGRID = SyncMapFormatTextGrid.DEFAULT
    """
    Alias for ``TEXTGRID_LONG``.
    """

    TEXTGRID_LONG = SyncMapFormatTextGrid.LONG
    """
    Praat full TextGrid format::

        File type = "ooTextFile"
        Object class = "TextGrid"

        xmin = 0.0
        xmax = 7.89
        tiers? <exists>
        size = 1
        item []:
            item [1]:
                class = "IntervalTier"
                name = "Token"
                xmin = 0.0
                xmax = 7.89
                intervals: size = 3
                intervals [1]:
                    xmin = 0.0
                    xmax = 1.234
                    text = "First fragment text"
                intervals [2]:
                    xmin = 1.234
                    xmax = 5.678
                    text = "Second fragment text"
                intervals [3]:
                    xmin = 5.678
                    xmax = 7.89
                    text = "Third fragment text"

    * Multiple levels: no (not yet)
    * Multiple lines: no

    See also http://www.fon.hum.uva.nl/praat/manual/TextGrid_file_formats.html

    Note that at the moment reading support is limited
    to the first tier in the TextGrid file.

    .. versionadded:: 1.7.0
    """

    TEXTGRID_SHORT = SyncMapFormatTextGrid.SHORT
    """
    Praat short TextGrid format::

        File type = "ooTextFile"
        Object class = "TextGrid"

        0.0
        7.89
        <exists>
        1
        "IntervalTier"
        "Token"
        0.0
        7.89
        3
        0.0
        1.234
        "First fragment text"
        1.234
        5.678
        "Second fragment text"
        5.678
        7.89
        "Third fragment text"

    * Multiple levels: no (not yet)
    * Multiple lines: no

    See also http://www.fon.hum.uva.nl/praat/manual/TextGrid_file_formats.html

    Note that at the moment reading support is limited
    to the first tier in the TextGrid file.

    .. versionadded:: 1.7.0
    """

    TSV = SyncMapFormatTSV.DEFAULT
    """
    Alias for ``TSVM``.
    """

    TSVH = SyncMapFormatTSV.HUMAN
    """
    Tab-separated plain text,
    with human-readable time values::

        00:00:00.000   00:00:01.234   f001
        00:00:01.234   00:00:05.678   f002
        00:00:05.678   00:00:07.890   f003

    * Multiple levels: no
    * Multiple lines: no

    .. versionadded:: 1.0.4
    """

    TSVM = SyncMapFormatTSV.MACHINE
    """
    Tab-separated plain text,
    with machine-readable time values,
    compatible with ``Audacity``::

        0.000   1.234   f001
        1.234   5.678   f002
        5.678   7.890   f003

    * Multiple levels: no
    * Multiple lines: no

    .. versionadded:: 1.2.0
    """

    TTML = SyncMapFormatTTML.DEFAULT
    """
    TTML caption/subtitle format
    (it might have multiple lines per fragment)::

        <?xml version="1.0" encoding="UTF-8" ?>
        <tt xmlns="http://www.w3.org/ns/ttml">
         <body>
          <div>
           <p xml:id="f001" begin="0.000" end="1.234">
            First fragment text
           </p>
           <p xml:id="f002" begin="1.234" end="5.678">
            Second fragment text<br/>Second line of second fragment
           </p>
           <p xml:id="f003" begin="5.678" end="7.890">
            Third fragment text<br/>Second line of third fragment
           </p>
          </div>
         </body>
        </tt>

    See also https://www.w3.org/TR/ttml1/

    * Multiple levels: yes (output only)
    * Multiple lines: yes
    """

    TXT = SyncMapFormatTXT.DEFAULT
    """
    Alias for ``TXTM``.
    """

    TXTH = SyncMapFormatTXT.HUMAN
    """
    Space-separated plain text
    with human-readable time values::

        f001 00:00:00.000 00:00:01.234 "First fragment text"
        f002 00:00:01.234 00:00:05.678 "Second fragment text"
        f003 00:00:05.678 00:00:07.890 "Third fragment text"

    * Multiple levels: no
    * Multiple lines: no

    Please note that the text is assumed to be contained
    in double quotes ("..."),
    which are stripped when reading from file,
    and added back when writing to file.

    .. versionadded:: 1.0.4
    """

    TXTM = SyncMapFormatTXT.MACHINE
    """
    Space-separated plain text,
    with machine-readable time values,
    compatible with ``SonicVisualizer``::

        f001 0.000 1.234 "First fragment text"
        f002 1.234 5.678 "Second fragment text"
        f003 5.678 7.890 "Third fragment text"

    * Multiple levels: no
    * Multiple lines: no

    Please note that the text is assumed to be contained
    in double quotes ("..."),
    which are stripped when reading from file,
    and added back when writing to file.

    .. versionadded:: 1.2.0
    """

    VTT = SyncMapFormatVTT.DEFAULT
    """
    WebVTT caption/subtitle format::

        WEBVTT

        1
        00:00:00.000 --> 00:00:01.234
        First fragment text

        2
        00:00:01.234 --> 00:00:05.678
        Second fragment text
        Second line of second fragment

        3
        00:00:05.678 --> 00:00:07.890
        Third fragment text
        Second line of third fragment

    * Multiple levels: no
    * Multiple lines: yes

    See also https://w3c.github.io/webvtt/

    Note that WebVTT files using tabs as separators
    cannot be read at the moment.
    Use spaces instead or pre-process your files,
    replacing tabs with spaces.
    """

    XML = SyncMapFormatXML.DEFAULT
    """
    XML::

        <?xml version="1.0" encoding="UTF-8" ?>
        <map>
         <fragment id="f001" begin="0.000" end="1.234">
          <line>First fragment text</line>
          <children></children>
         </fragment>
         <fragment id="f002" begin="1.234" end="5.678">
          <line>Second fragment text</line>
          <line>Second line of second fragment</line>
          <children></children>
         </fragment>
         <fragment id="f003" begin="5.678" end="7.890">
          <line>Third fragment text</line>
          <line>Second line of third fragment</line>
          <children></children>
         </fragment>
        </map>

    * Multiple levels: yes (output only)
    * Multiple lines: yes
    """

    XML_LEGACY = SyncMapFormatXMLLegacy.DEFAULT
    """
    XML, legacy format::

        <?xml version="1.0" encoding="UTF-8" ?>
        <map>
         <fragment>
          <identifier>f001</identifier>
          <start>0.000</start>
          <end>1.234</end>
         </fragment>
         <fragment>
          <identifier>f002</identifier>
          <start>1.234</start>
          <end>5.678</end>
         </fragment>
         <fragment>
          <identifier>f003</identifier>
          <start>5.678</start>
          <end>7.890</end>
         </fragment>
        </map>

    * Multiple levels: no
    * Multiple lines: no

    Deprecated, it will be removed in v2.0.0.
    Use ``XML`` instead.

    .. deprecated:: 1.2.0
    """

    CODE_TO_CLASS = {
        AUD: SyncMapFormatAudacity,
        AUDH: SyncMapFormatAudacity,
        AUDM: SyncMapFormatAudacity,
        CSV: SyncMapFormatCSV,
        CSVH: SyncMapFormatCSV,
        CSVM: SyncMapFormatCSV,
        DFXP: SyncMapFormatTTML,
        EAF: SyncMapFormatEAF,
        JSON: SyncMapFormatJSON,
        RBSE: SyncMapFormatRBSE,
        SBV: SyncMapFormatSUB,
        SMIL: SyncMapFormatSMIL,
        SMILH: SyncMapFormatSMIL,
        SMILM: SyncMapFormatSMIL,
        SRT: SyncMapFormatSRT,
        SSV: SyncMapFormatSSV,
        SSVH: SyncMapFormatSSV,
        SSVM: SyncMapFormatSSV,
        SUB: SyncMapFormatSUB,
        TAB: SyncMapFormatTSV,
        TEXTGRID: SyncMapFormatTextGrid,
        TEXTGRID_LONG: SyncMapFormatTextGrid,
        TEXTGRID_SHORT: SyncMapFormatTextGrid,
        TSV: SyncMapFormatTSV,
        TSVH: SyncMapFormatTSV,
        TSVM: SyncMapFormatTSV,
        TTML: SyncMapFormatTTML,
        TXT: SyncMapFormatTXT,
        TXTH: SyncMapFormatTXT,
        TXTM: SyncMapFormatTXT,
        VTT: SyncMapFormatVTT,
        XML: SyncMapFormatXML,
        XML_LEGACY: SyncMapFormatXMLLegacy,
    }

    ALLOWED_VALUES = sorted(list(CODE_TO_CLASS.keys()))
    """ List of all the allowed values """



#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the following classes:

* :class:`~aeneas.audiofile.AudioFile`, representing an audio file;
* :class:`~aeneas.audiofile.AudioFileConverterError`,
* :class:`~aeneas.audiofile.AudioFileNotInitializedError`,
* :class:`~aeneas.audiofile.AudioFileProbeError`, and
* :class:`~aeneas.audiofile.AudioFileUnsupportedFormatError`,
  representing errors generated by audio files.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy

from aeneas.exacttiming import TimeValue
from aeneas.ffmpegwrapper import FFMPEGPathError
from aeneas.ffmpegwrapper import FFMPEGWrapper
from aeneas.ffprobewrapper import FFPROBEParsingError
from aeneas.ffprobewrapper import FFPROBEPathError
from aeneas.ffprobewrapper import FFPROBEUnsupportedFormatError
from aeneas.ffprobewrapper import FFPROBEWrapper
from aeneas.logger import Loggable
from aeneas.runtimeconfiguration import RuntimeConfiguration
from aeneas.wavfile import read as scipywavread
from aeneas.wavfile import write as scipywavwrite
import aeneas.globalfunctions as gf


class AudioFileConverterError(Exception):
    """
    Error raised when the audio converter executable cannot be executed.
    """
    pass


class AudioFileNotInitializedError(Exception):
    """
    Error raised when trying to access audio samples from
    an :class:`~aeneas.audiofile.AudioFile` object which
    has not been initialized yet.
    """
    pass


class AudioFileProbeError(Exception):
    """
    Error raised when the audio probe executable cannot be executed.
    """
    pass


class AudioFileUnsupportedFormatError(Exception):
    """
    Error raised when the format of the given file cannot be decoded.
    """
    pass


class AudioFile(Loggable):
    """
    A class representing an audio file.

    This class can be used either to extract properties
    from an audio file on disk,
    or to load/edit/save a monoaural (single channel) audio file,
    represented as an array of audio samples.

    The properties of the audio file (length, format, etc.)
    can set by invoking the :func:`~aeneas.audiofile.AudioFile.read_properties` function,
    which calls an audio file probe.
    (Currently, the probe is :class:`~aeneas.ffprobewrapper.FFPROBEWrapper`)

    Moreover, this class can read the audio data,
    by converting the original file format
    into a temporary PCM16 Mono WAVE (RIFF) file,
    which is deleted as soon as audio data is read in memory.
    (Currently, the converter is :class:`~aeneas.ffmpegwrapper.FFMPEGWrapper`)

    The internal representation of the wave is a
    a NumPy 1D array of ``float64`` values in ``[-1.0, 1.0]``.
    It supports append, reverse, and trim operations.
    Audio samples can be written to file.
    Memory can be pre-allocated to speed append operations up.
    Allocated memory is doubled when an append operation
    requires more memory than what is available;
    this leads to an amortized linear complexity
    (in the number of audio samples)
    for append operations.

    .. note:: Support for stereo WAVE files might be implemented in a future version

    :param string file_path: the path of the audio file
    :param tuple file_format: the format of the audio file, if known in advance: ``(codec, channels, rate)`` or ``None``
    :param rconf: a runtime configuration
    :type  rconf: :class:`~aeneas.runtimeconfiguration.RuntimeConfiguration`
    :param logger: the logger object
    :type  logger: :class:`~aeneas.logger.Logger`
    """

    FILE_EXTENSIONS = [
        u"3g2",
        u"3gp",
        u"aa",
        u"aa3",
        u"aac",
        u"aax",
        u"aiff",
        u"alac",
        u"amr",
        u"ape",
        u"asf",
        u"at3",
        u"at9",
        u"au",
        u"avi",
        u"awb",
        u"celt",
        u"dct",
        u"dss",
        u"dvf",
        u"eac",
        u"flac",
        u"flv",
        u"gsm",
        u"m4a",
        u"m4b",
        u"m4p",
        u"m4v",
        u"mid",
        u"midi",
        u"mkv",
        u"mmf",
        u"mov",
        u"mp2",
        u"mp3",
        u"mp4",
        u"mpc",
        u"mpeg",
        u"mpg",
        u"mpv",
        u"msv",
        u"oga",
        u"ogg",
        u"ogv",
        u"oma",
        u"opus",
        u"pcm",
        u"qt",
        u"ra",
        u"ram",
        u"raw",
        u"riff",
        u"rm",
        u"rmvb",
        u"shn",
        u"sln",
        u"theora",
        u"tta",
        u"vob",
        u"vorbis",
        u"vox",
        u"wav",
        u"webm",
        u"wma",
        u"wmv",
        u"wv",
        u"yuv",
    ]
    """ Extensions of common formats for audio (and video) files. """

    TAG = u"AudioFile"

    def __init__(self, file_path=None, file_format=None, rconf=None, logger=None):
        super(AudioFile, self).__init__(rconf=rconf, logger=logger)
        self.file_path = file_path
        self.file_format = file_format
        self.file_size = None
        self.audio_length = None
        self.audio_format = None
        self.audio_sample_rate = None
        self.audio_channels = None
        self.__samples_capacity = 0
        self.__samples_length = 0
        self.__samples = None

    def __unicode__(self):
        fmt = self.file_format
        if isinstance(fmt, tuple):
            fmt = u"%s %d %d" % fmt
        msg = [
            u"File path:         %s" % self.file_path,
            u"File format:       %s" % fmt,
            u"File size (bytes): %s" % gf.safe_int(self.file_size),
            u"Audio length (s):  %s" % gf.safe_float(self.audio_length),
            u"Audio format:      %s" % self.audio_format,
            u"Audio sample rate: %s" % gf.safe_int(self.audio_sample_rate),
            u"Audio channels:    %s" % gf.safe_int(self.audio_channels),
            u"Samples capacity:  %s" % gf.safe_int(self.__samples_capacity),
            u"Samples length:    %s" % gf.safe_int(self.__samples_length),
        ]
        return u"\n".join(msg)

    def __str__(self):
        return gf.safe_str(self.__unicode__())

    @property
    def file_path(self):
        """
        The path of the audio file.

        :rtype: string
        """
        return self.__file_path

    @file_path.setter
    def file_path(self, file_path):
        self.__file_path = file_path

    @property
    def file_size(self):
        """
        The size of the audio file, in bytes.

        :rtype: int
        """
        return self.__file_size

    @file_size.setter
    def file_size(self, file_size):
        self.__file_size = file_size

    @property
    def audio_length(self):
        """
        The length of the audio file, in seconds.

        :rtype: :class:`~aeneas.exacttiming.TimeValue`
        """
        return self.__audio_length

    @audio_length.setter
    def audio_length(self, audio_length):
        self.__audio_length = audio_length

    @property
    def audio_format(self):
        """
        The format of the audio file.

        :rtype: string
        """
        return self.__audio_format

    @audio_format.setter
    def audio_format(self, audio_format):
        self.__audio_format = audio_format

    @property
    def audio_sample_rate(self):
        """
        The sample rate of the audio file, in samples per second.

        :rtype: int
        """
        return self.__audio_sample_rate

    @audio_sample_rate.setter
    def audio_sample_rate(self, audio_sample_rate):
        self.__audio_sample_rate = audio_sample_rate

    @property
    def audio_channels(self):
        """
        The number of channels of the audio file.

        :rtype: int
        """
        return self.__audio_channels

    @audio_channels.setter
    def audio_channels(self, audio_channels):
        self.__audio_channels = audio_channels

    @property
    def audio_samples(self):
        """
        The audio audio_samples, that is, an array of ``float64`` values,
        each representing an audio sample in ``[-1.0, 1.0]``.

        Note that this function returns a view into the
        first ``self.__samples_length`` elements of ``self.__samples``.
        If you want to clone the values,
        you must use e.g. ``numpy.array(audiofile.audio_samples)``.

        :rtype: :class:`numpy.ndarray` (1D, view)
        :raises: :class:`~aeneas.audiofile.AudioFileNotInitializedError`: if the audio file is not initialized yet
        """
        if self.__samples is None:
            if self.file_path is None:
                self.log_exc(u"AudioFile object not initialized", None, True, AudioFileNotInitializedError)
            else:
                self.read_samples_from_file()
        return self.__samples[0:self.__samples_length]

    def read_properties(self):
        """
        Populate this object by reading
        the audio properties of the file at the given path.

        Currently this function uses
        :class:`~aeneas.ffprobewrapper.FFPROBEWrapper`
        to get the audio file properties.

        :raises: :class:`~aeneas.audiofile.AudioFileProbeError`: if the path to the ``ffprobe`` executable cannot be called
        :raises: :class:`~aeneas.audiofile.AudioFileUnsupportedFormatError`: if the audio file has a format not supported
        :raises: OSError: if the audio file cannot be read
        """
        self.log(u"Reading properties...")

        # check the file can be read
        if not gf.file_can_be_read(self.file_path):
            self.log_exc(u"File '%s' cannot be read" % (self.file_path), None, True, OSError)

        # get the file size
        self.log([u"Getting file size for '%s'", self.file_path])
        self.file_size = gf.file_size(self.file_path)
        self.log([u"File size for '%s' is '%d'", self.file_path, self.file_size])

        # get the audio properties using FFPROBEWrapper
        try:
            self.log(u"Reading properties with FFPROBEWrapper...")
            properties = FFPROBEWrapper(
                rconf=self.rconf,
                logger=self.logger
            ).read_properties(self.file_path)
            self.log(u"Reading properties with FFPROBEWrapper... done")
        except FFPROBEPathError:
            self.log_exc(u"Unable to call ffprobe executable", None, True, AudioFileProbeError)
        except (FFPROBEUnsupportedFormatError, FFPROBEParsingError):
            self.log_exc(u"Audio file format not supported by ffprobe", None, True, AudioFileUnsupportedFormatError)

        # save relevant properties in results inside the audiofile object
        self.audio_length = TimeValue(properties[FFPROBEWrapper.STDOUT_DURATION])
        self.audio_format = properties[FFPROBEWrapper.STDOUT_CODEC_NAME]
        self.audio_sample_rate = gf.safe_int(properties[FFPROBEWrapper.STDOUT_SAMPLE_RATE])
        self.audio_channels = gf.safe_int(properties[FFPROBEWrapper.STDOUT_CHANNELS])
        self.log([u"Stored audio_length: '%s'", self.audio_length])
        self.log([u"Stored audio_format: '%s'", self.audio_format])
        self.log([u"Stored audio_sample_rate: '%s'", self.audio_sample_rate])
        self.log([u"Stored audio_channels: '%s'", self.audio_channels])
        self.log(u"Reading properties... done")

    def read_samples_from_file(self):
        """
        Load the audio samples from file into memory.

        If ``self.file_format`` is ``None`` or it is not
        ``("pcm_s16le", 1, self.rconf.sample_rate)``,
        the file will be first converted
        to a temporary PCM16 mono WAVE file.
        Audio data will be read from this temporary file,
        which will be then deleted from disk immediately.

        Otherwise,
        the audio data will be read directly
        from the given file,
        which will not be deleted from disk.

        :raises: :class:`~aeneas.audiofile.AudioFileConverterError`: if the path to the ``ffmpeg`` executable cannot be called
        :raises: :class:`~aeneas.audiofile.AudioFileUnsupportedFormatError`: if the audio file has a format not supported
        :raises: OSError: if the audio file cannot be read
        """
        self.log(u"Loading audio data...")

        # check the file can be read
        if not gf.file_can_be_read(self.file_path):
            self.log_exc(u"File '%s' cannot be read" % (self.file_path), None, True, OSError)

        # determine if we need to convert the audio file
        convert_audio_file = (
            (self.file_format is None) or
            (
                (self.rconf.safety_checks) and
                (self.file_format != ("pcm_s16le", 1, self.rconf.sample_rate))
            )
        )

        # convert the audio file if needed
        if convert_audio_file:
            # convert file to PCM16 mono WAVE with correct sample rate
            self.log(u"self.file_format is None or not good => converting self.file_path")
            tmp_handler, tmp_file_path = gf.tmp_file(suffix=u".wav", root=self.rconf[RuntimeConfiguration.TMP_PATH])
            self.log([u"Temporary PCM16 mono WAVE file: '%s'", tmp_file_path])
            try:
                self.log(u"Converting audio file to mono...")
                converter = FFMPEGWrapper(rconf=self.rconf, logger=self.logger)
                converter.convert(self.file_path, tmp_file_path)
                self.file_format = ("pcm_s16le", 1, self.rconf.sample_rate)
                self.log(u"Converting audio file to mono... done")
            except FFMPEGPathError:
                gf.delete_file(tmp_handler, tmp_file_path)
                self.log_exc(u"Unable to call ffmpeg executable", None, True, AudioFileConverterError)
            except OSError:
                gf.delete_file(tmp_handler, tmp_file_path)
                self.log_exc(u"Audio file format not supported by ffmpeg", None, True, AudioFileUnsupportedFormatError)
        else:
            # read the file directly
            if self.rconf.safety_checks:
                self.log(u"self.file_format is good => reading self.file_path directly")
            else:
                self.log_warn(u"Safety checks disabled => reading self.file_path directly")
            tmp_handler = None
            tmp_file_path = self.file_path

        # TODO allow calling C extension cwave to read samples faster
        try:
            self.audio_format = "pcm16"
            self.audio_channels = 1
            self.audio_sample_rate, self.__samples = scipywavread(tmp_file_path)
            # scipy reads a sample as an int16_t, that is, a number in [-32768, 32767]
            # so we convert it to a float64 in [-1, 1]
            self.__samples = self.__samples.astype("float64") / 32768
            self.__samples_capacity = len(self.__samples)
            self.__samples_length = self.__samples_capacity
            self._update_length()
        except ValueError:
            self.log_exc(u"Audio format not supported by scipywavread", None, True, AudioFileUnsupportedFormatError)

        # if we converted the audio file, delete the temporary converted audio file
        if convert_audio_file:
            gf.delete_file(tmp_handler, tmp_file_path)
            self.log([u"Deleted temporary audio file: '%s'", tmp_file_path])

        self._update_length()
        self.log([u"Sample length:  %.3f", self.audio_length])
        self.log([u"Sample rate:    %d", self.audio_sample_rate])
        self.log([u"Audio format:   %s", self.audio_format])
        self.log([u"Audio channels: %d", self.audio_channels])
        self.log(u"Loading audio data... done")

    def preallocate_memory(self, capacity):
        """
        Preallocate memory to store audio samples,
        to avoid repeated new allocations and copies
        while performing several consecutive append operations.

        If ``self.__samples`` is not initialized,
        it will become an array of ``capacity`` zeros.

        If ``capacity`` is larger than the current capacity,
        the current ``self.__samples`` will be extended with zeros.

        If ``capacity`` is smaller than the current capacity,
        the first ``capacity`` values of ``self.__samples``
        will be retained.

        :param int capacity: the new capacity, in number of samples
        :raises: ValueError: if ``capacity`` is negative

        .. versionadded:: 1.5.0
        """
        if capacity < 0:
            raise ValueError(u"The capacity value cannot be negative")
        if self.__samples is None:
            self.log(u"Not initialized")
            self.__samples = numpy.zeros(capacity)
            self.__samples_length = 0
        else:
            self.log([u"Previous sample length was   (samples): %d", self.__samples_length])
            self.log([u"Previous sample capacity was (samples): %d", self.__samples_capacity])
            self.__samples = numpy.resize(self.__samples, capacity)
            self.__samples_length = min(self.__samples_length, capacity)
        self.__samples_capacity = capacity
        self.log([u"Current sample capacity is   (samples): %d", self.__samples_capacity])

    def minimize_memory(self):
        """
        Reduce the allocated memory to the minimum
        required to store the current audio samples.

        This function is meant to be called
        when building a wave incrementally,
        after the last append operation.

        .. versionadded:: 1.5.0
        """
        if self.__samples is None:
            self.log(u"Not initialized, returning")
        else:
            self.log(u"Initialized, minimizing memory...")
            self.preallocate_memory(self.__samples_length)
            self.log(u"Initialized, minimizing memory... done")

    def add_samples(self, samples, reverse=False):
        """
        Concatenate the given new samples to the current audio data.

        This function initializes the memory if no audio data
        is present already.

        If ``reverse`` is ``True``, the new samples
        will be reversed and then concatenated.

        :param samples: the new samples to be concatenated
        :type  samples: :class:`numpy.ndarray` (1D)
        :param bool reverse: if ``True``, concatenate new samples after reversing them

        .. versionadded:: 1.2.1
        """
        self.log(u"Adding samples...")
        samples_length = len(samples)
        current_length = self.__samples_length
        future_length = current_length + samples_length
        if (self.__samples is None) or (self.__samples_capacity < future_length):
            self.preallocate_memory(2 * future_length)
        if reverse:
            self.__samples[current_length:future_length] = samples[::-1]
        else:
            self.__samples[current_length:future_length] = samples[:]
        self.__samples_length = future_length
        self._update_length()
        self.log(u"Adding samples... done")

    def reverse(self):
        """
        Reverse the audio data.

        :raises: :class:`~aeneas.audiofile.AudioFileNotInitializedError`: if the audio file is not initialized yet

        .. versionadded:: 1.2.0
        """
        if self.__samples is None:
            if self.file_path is None:
                self.log_exc(u"AudioFile object not initialized", None, True, AudioFileNotInitializedError)
            else:
                self.read_samples_from_file()
        self.log(u"Reversing...")
        self.__samples[0:self.__samples_length] = numpy.flipud(self.__samples[0:self.__samples_length])
        self.log(u"Reversing... done")

    def trim(self, begin=None, length=None):
        """
        Get a slice of the audio data of ``length`` seconds,
        starting from ``begin`` seconds.

        If audio data is not loaded, load it and then slice it.

        :param begin: the start position, in seconds
        :type  begin: :class:`~aeneas.exacttiming.TimeValue`
        :param length: the  position, in seconds
        :type  length: :class:`~aeneas.exacttiming.TimeValue`
        :raises: TypeError: if one of the arguments is not ``None``
                            or :class:`~aeneas.exacttiming.TimeValue`

        .. versionadded:: 1.2.0
        """
        for variable, name in [(begin, "begin"), (length, "length")]:
            if (variable is not None) and (not isinstance(variable, TimeValue)):
                raise TypeError(u"%s is not None or TimeValue" % name)
        self.log(u"Trimming...")
        if (begin is None) and (length is None):
            self.log(u"begin and length are both None: nothing to do")
        else:
            if begin is None:
                begin = TimeValue("0.000")
                self.log([u"begin was None, now set to %.3f", begin])
            begin = min(max(TimeValue("0.000"), begin), self.audio_length)
            self.log([u"begin is %.3f", begin])
            if length is None:
                length = self.audio_length - begin
                self.log([u"length was None, now set to %.3f", length])
            length = min(max(TimeValue("0.000"), length), self.audio_length - begin)
            self.log([u"length is %.3f", length])
            begin_index = int(begin * self.audio_sample_rate)
            end_index = int((begin + length) * self.audio_sample_rate)
            new_idx = end_index - begin_index
            self.__samples[0:new_idx] = self.__samples[begin_index:end_index]
            self.__samples_length = new_idx
            self._update_length()
        self.log(u"Trimming... done")

    def write(self, file_path):
        """
        Write the audio data to file.
        Return ``True`` on success, or ``False`` otherwise.

        :param string file_path: the path of the output file to be written
        :raises: :class:`~aeneas.audiofile.AudioFileNotInitializedError`: if the audio file is not initialized yet

        .. versionadded:: 1.2.0
        """
        if self.__samples is None:
            if self.file_path is None:
                self.log_exc(u"AudioFile object not initialized", None, True, AudioFileNotInitializedError)
            else:
                self.read_samples_from_file()
        self.log([u"Writing audio file '%s'...", file_path])
        try:
            # our value is a float64 in [-1, 1]
            # scipy writes the sample as an int16_t, that is, a number in [-32768, 32767]
            data = (self.audio_samples * 32768).astype("int16")
            scipywavwrite(file_path, self.audio_sample_rate, data)
        except Exception as exc:
            self.log_exc(u"Error writing audio file to '%s'" % (file_path), exc, True, OSError)
        self.log([u"Writing audio file '%s'... done", file_path])

    def clear_data(self):
        """
        Clear the audio data, freeing memory.
        """
        self.log(u"Clear audio_data")
        self.__samples_capacity = 0
        self.__samples_length = 0
        self.__samples = None

    def _update_length(self):
        """
        Update the audio length property,
        according to the length of the current audio data
        and audio sample rate.

        This function fails silently if one of the two is ``None``.
        """
        if (self.audio_sample_rate is not None) and (self.__samples is not None):
            # NOTE computing TimeValue (... / ...) yields wrong results,
            #      see issue #168
            #      self.audio_length = TimeValue(self.__samples_length / self.audio_sample_rate)
            self.audio_length = TimeValue(self.__samples_length) / TimeValue(self.audio_sample_rate)



#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Set the aeneas package up.
"""

from setuptools import Extension
from setuptools import setup
import io
import os
import shutil
import sys

from setupmeta import PKG_AUTHOR
from setupmeta import PKG_AUTHOR_EMAIL
from setupmeta import PKG_CLASSIFIERS
from setupmeta import PKG_EXTRAS_REQUIRE
from setupmeta import PKG_INSTALL_REQUIRES
from setupmeta import PKG_KEYWORDS
from setupmeta import PKG_LICENSE
from setupmeta import PKG_LONG_DESCRIPTION
from setupmeta import PKG_NAME
from setupmeta import PKG_PACKAGES
from setupmeta import PKG_PACKAGE_DATA
from setupmeta import PKG_SCRIPTS
from setupmeta import PKG_SHORT_DESCRIPTION
from setupmeta import PKG_URL
from setupmeta import PKG_VERSION

__author__ = "Alberto Pettarin"
__email__ = "aeneas@readbeyond.it"
__copyright__ = """
    Copyright 2012-2013, Alberto Pettarin (www.albertopettarin.it)
    Copyright 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
    Copyright 2015-2017, Alberto Pettarin (www.albertopettarin.it)
"""
__license__ = "GNU AGPL 3"
__status__ = "Production"
__version__ = "1.7.3"


##############################################################################
#
# Windows-specific setup for compiling cew
#
##############################################################################

def prepare_cew_for_windows():
    """
    Copy files needed to compile the ``cew`` Python C extension on Windows.

    A glorious day, when Microsoft will offer a decent support
    for Python and shared libraries,
    all this mess will be unnecessary and it should be removed.
    May that day come soon.

    Return ``True`` if successful, ``False`` otherwise.

    :rtype: bool
    """
    try:
        # copy espeak_sapi.dll to C:\Windows\System32\espeak.dll
        espeak_dll_win_path = "C:\\Windows\\System32\\espeak.dll"
        espeak_dll_dst_path = "aeneas\\cew\\espeak.dll"
        espeak_dll_src_paths = [
            "C:\\aeneas\\eSpeak\\espeak_sapi.dll",
            "C:\\sync\\eSpeak\\espeak_sapi.dll",
            "C:\\Program Files\\eSpeak\\espeak_sapi.dll",
            "C:\\Program Files (x86)\\eSpeak\\espeak_sapi.dll",
        ]
        if os.path.exists(espeak_dll_dst_path):
            print("[INFO] Found eSpeak DLL in %s" % espeak_dll_dst_path)
        else:
            found = False
            copied = False
            for src_path in espeak_dll_src_paths:
                if os.path.exists(src_path):
                    found = True
                    print("[INFO] Copying eSpeak DLL from %s into %s" % (src_path, espeak_dll_dst_path))
                    try:
                        shutil.copyfile(src_path, espeak_dll_dst_path)
                        copied = True
                        print("[INFO] Copied eSpeak DLL")
                    except:
                        pass
                    break
            if not found:
                print("[WARN] Unable to find the eSpeak DLL, probably because you installed eSpeak in a non-standard location.")
                print("[WARN] If you want to run aeneas with the C extension cew,")
                print("[WARN] please copy espeak_sapi.dll from your eSpeak directory to %s" % espeak_dll_win_path)
                # print("[WARN] and run the aeneas setup again.")
                # return False
            elif not copied:
                print("[WARN] Unable to copy the eSpeak DLL, probably because you are not running with admin privileges.")
                print("[WARN] If you want to run aeneas with the C extension cew,")
                print("[WARN] please copy espeak_sapi.dll from your eSpeak directory to %s" % espeak_dll_win_path)
                # print("[WARN] and run the aeneas setup again.")
                # return False

        # NOTE: espeak.lib is needed only while compiling the C extension, not when using it
        #       so, we copy it in the current working directory from the included thirdparty\ directory
        # NOTE: PREV: copy thirdparty\espeak.lib to $PYTHON\libs\espeak.lib
        # NOTE: PREV: espeak_lib_dst_path = os.path.join(sys.prefix, "libs", "espeak.lib")
        espeak_lib_src_path = os.path.join(os.path.dirname(__file__), "thirdparty", "espeak.lib")
        espeak_lib_dst_path = os.path.join(os.path.dirname(__file__), "espeak.lib")
        if os.path.exists(espeak_lib_dst_path):
            print("[INFO] Found eSpeak LIB in %s" % espeak_lib_dst_path)
        else:
            try:
                print("[INFO] Copying eSpeak LIB into %s" % espeak_lib_dst_path)
                shutil.copyfile(espeak_lib_src_path, espeak_lib_dst_path)
                print("[INFO] Copied eSpeak LIB")
            except:
                print("[WARN] Unable to copy the eSpeak LIB, probably because you are not running with admin privileges.")
                print("[WARN] If you want to compile the C extension cew,")
                print("[WARN] please copy espeak.lib from the thirdparty directory into %s" % espeak_lib_dst_path)
                print("[WARN] and run the aeneas setup again.")
                return False

        # if here, we have completed the setup, return True
        return True
    except Exception as e:
        print("[WARN] Unexpected exception while preparing cew: %s" % e)
    return False


##############################################################################
#
# find the OS out and read environment variables for options
#
##############################################################################

# get platform
IS_LINUX = (os.name == "posix") and (os.uname()[0] == "Linux")
IS_OSX = (os.name == "posix") and (os.uname()[0] == "Darwin")
IS_WINDOWS = (os.name == "nt")

# define what values of environment variables are considered equal to True
TRUE_VALUES = [
    "TRUE",
    "True",
    "true",
    "YES",
    "Yes",
    "yes",
    "1",
    1
]

# check whether the user set additional parameters using environment variables
# NOTE by the book this should be done by subclassing the setuptools Distribution object
#      but for now using environment variables is good enough
WITHOUT_CDTW = os.getenv("AENEAS_WITH_CDTW", "True") not in TRUE_VALUES
WITHOUT_CMFCC = os.getenv("AENEAS_WITH_CMFCC", "True") not in TRUE_VALUES
WITHOUT_CEW = os.getenv("AENEAS_WITH_CEW", "True") not in TRUE_VALUES
FORCE_CEW = os.getenv("AENEAS_FORCE_CEW", "False") in TRUE_VALUES
FORCE_CFW = os.getenv("AENEAS_FORCE_CFW", "False") in TRUE_VALUES


##############################################################################
#
# actual setup
#
##############################################################################

# try importing numpy: if it fails, warn user and exit
try:
    from numpy import get_include
    from numpy.distutils import misc_util
except ImportError:
    print("[ERRO] You must install numpy before installing aeneas")
    print("[INFO] Try the following command:")
    print("[INFO] $ sudo pip install numpy")
    sys.exit(1)

# to compile cdtw and cmfcc, we need to include the NumPy dirs
INCLUDE_DIRS = [misc_util.get_numpy_include_dirs()]

# scripts to be installed globally
# on Linux and Mac OS X, use the file without extension
# on Windows, use the file with .py extension
if IS_WINDOWS:
    PKG_SCRIPTS = [s + ".py" for s in PKG_SCRIPTS]

# prepare Extension objects
EXTENSION_CDTW = Extension(
    name="aeneas.cdtw.cdtw",
    sources=[
        "aeneas/cdtw/cdtw_py.c",
        "aeneas/cdtw/cdtw_func.c",
        "aeneas/cint/cint.c"
    ],
    include_dirs=[
        get_include()
    ]
)
EXTENSION_CMFCC = Extension(
    name="aeneas.cmfcc.cmfcc",
    sources=[
        "aeneas/cmfcc/cmfcc_py.c",
        "aeneas/cmfcc/cmfcc_func.c",
        "aeneas/cwave/cwave_func.c",
        "aeneas/cint/cint.c"
    ],
    include_dirs=[
        get_include()
    ]
)
EXTENSION_CEW = Extension(
    name="aeneas.cew.cew",
    sources=[
        "aeneas/cew/cew_py.c",
        "aeneas/cew/cew_func.c"
    ],
    libraries=[
        "espeak"
    ]
)
EXTENSION_CFW = Extension(
    name="aeneas.cfw.cfw",
    sources=[
        "aeneas/cfw/cfw_py.cc",
        "aeneas/cfw/cfw_func.cc"
    ],
    include_dirs=[
        "aeneas/cfw/festival",
        "aeneas/cfw/speech_tools"
    ],
    libraries=[
        "Festival",
        "estools",
        "estbase",
        "eststring",
    ]
)
# cwave is ready, but currently not used
# EXTENSION_CWAVE = Extension(
#     name="aeneas.cwave.cwave",
#     sources=[
#         "aeneas/cwave/cwave_py.c",
#         "aeneas/cwave/cwave_func.c"
#     ],
#     include_dirs=[
#         get_include()
#     ]
# )

# append or ignore cew extension as requested
EXTENSIONS = []

if WITHOUT_CDTW:
    print("[INFO] ************************************************************")
    print("[INFO] The user specified AENEAS_WITH_CDTW=False: not building cdtw")
    print("[INFO] ************************************************************")
    print("[INFO] ")
else:
    EXTENSIONS.append(EXTENSION_CDTW)

if WITHOUT_CMFCC:
    print("[INFO] **************************************************************")
    print("[INFO] The user specified AENEAS_WITH_CMFCC=False: not building cmfcc")
    print("[INFO] **************************************************************")
    print("[INFO] ")
else:
    EXTENSIONS.append(EXTENSION_CMFCC)

if WITHOUT_CEW:
    print("[INFO] **********************************************************")
    print("[INFO] The user specified AENEAS_WITH_CEW=False: not building cew")
    print("[INFO] **********************************************************")
    print("[INFO] ")
elif FORCE_CEW:
    print("[INFO] ********************************************************************************")
    print("[INFO] The user specified AENEAS_FORCE_CEW=True: attempting to build cew without checks")
    print("[INFO] ********************************************************************************")
    print("[INFO] ")
    EXTENSIONS.append(EXTENSION_CEW)
else:
    if IS_LINUX:
        EXTENSIONS.append(EXTENSION_CEW)
    elif IS_OSX:
        print("[INFO] *********************************************************************************")
        print("[INFO] Compiling the C extension cew on Mac OS X is experimental.")
        print("[INFO] ")
        print("[INFO] Before installing aeneas with cew, you must run:")
        print("[INFO] $ brew update && brew upgrade --cleanup espeak")
        print("[INFO] to run the new brew formula installing libespeak, the library version of espeak.")
        print("[INFO] ")
        print("[INFO] If you experience problems, disable cew compilation by specifying")
        print("[INFO] the environment variable AENEAS_WITH_CEW=False .")
        print("[INFO] Please see the aeneas installation documentation for details.")
        print("[INFO] ********************************************************************************")
        print("[INFO] ")
        EXTENSIONS.append(EXTENSION_CEW)
    elif IS_WINDOWS:
        print("[INFO] *****************************************************************")
        print("[INFO] Compiling the C extension cew on Windows is experimental.")
        print("[INFO] ")
        print("[INFO] If you experience problems, disable cew compilation by specifying")
        print("[INFO] the environment variable AENEAS_WITH_CEW=False .")
        print("[INFO] Please see the aeneas installation documentation for details.")
        print("[INFO] *****************************************************************")
        print("[INFO] ")
        if prepare_cew_for_windows():
            EXTENSIONS.append(EXTENSION_CEW)
        else:
            print("[WARN] Unable to complete the setup for C extension cew, not building it.")
    else:
        print("[INFO] The C extension cew is not available for your OS.")

if FORCE_CFW:
    print("[INFO] ********************************************************************************")
    print("[INFO] The user specified AENEAS_FORCE_CFW=True: attempting to build cfw without checks")
    print("[INFO] ********************************************************************************")
    print("[INFO] ")
    EXTENSIONS.append(EXTENSION_CFW)

# now we are ready to call setup()
setup(
    name=PKG_NAME,
    version=PKG_VERSION,
    packages=PKG_PACKAGES,
    package_data=PKG_PACKAGE_DATA,
    description=PKG_SHORT_DESCRIPTION,
    long_description=PKG_LONG_DESCRIPTION,
    author=PKG_AUTHOR,
    author_email=PKG_AUTHOR_EMAIL,
    url=PKG_URL,
    license=PKG_LICENSE,
    keywords=PKG_KEYWORDS,
    classifiers=PKG_CLASSIFIERS,
    install_requires=PKG_INSTALL_REQUIRES,
    extras_require=PKG_EXTRAS_REQUIRE,
    scripts=PKG_SCRIPTS,
    include_dirs=INCLUDE_DIRS,
    ext_modules=EXTENSIONS
)



#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the following classes:

* :class:`~aeneas.ttswrappers.nuancettswrapper.NuanceTTSWrapper`,
  a wrapper for the Nuance TTS API engine.

Please refer to
https://developer.nuance.com/
for further details.

.. note:: This module requires Python module ``requests`` (``pip install requests``).

.. warning:: You will be billed according to your Nuance Developers account plan.

.. warning:: This module is experimental, use at your own risk.

.. versionadded:: 1.5.0
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy
import time
import uuid

from aeneas.audiofile import AudioFile
from aeneas.exacttiming import TimeValue
from aeneas.language import Language
from aeneas.runtimeconfiguration import RuntimeConfiguration
from aeneas.ttswrappers.basettswrapper import BaseTTSWrapper
import aeneas.globalfunctions as gf


class NuanceTTSWrapper(BaseTTSWrapper):
    """
    A wrapper for the Nuance Developers TTS API.

    This wrapper supports calling the TTS engine
    only via Python.

    In abstract terms, it performs one or more calls to the
    Nuance TTS API service, and concatenate the resulting WAVE files,
    returning their anchor times.

    To use this TTS engine, specify ::

        "tts=nuance|nuance_tts_api_id=...|nuance_tts_api_key=..."

    in the ``RuntimeConfiguration`` object,
    substituting your Nuance Developer API ID and Key.

    You might also want to enable the TTS caching,
    to reduce the number of API calls ::

        "tts=nuance|tts_cache=True"

    See :class:`~aeneas.ttswrappers.basettswrapper.BaseTTSWrapper`
    for the available functions.
    Below are listed the languages supported by this wrapper.

    :param rconf: a runtime configuration
    :type  rconf: :class:`~aeneas.runtimeconfiguration.RuntimeConfiguration`
    :param logger: the logger object
    :type  logger: :class:`~aeneas.logger.Logger`
    """

    ARA = Language.ARA
    """ Arabic """

    CAT = Language.CAT
    """ Catalan """

    CES = Language.CES
    """ Czech """

    CMN = Language.CMN
    """ Mandarin Chinese """

    DAN = Language.DAN
    """ Danish """

    DEU = Language.DEU
    """ German """

    ELL = Language.ELL
    """ Greek (Modern) """

    ENG = Language.ENG
    """ English """

    EUS = Language.EUS
    """ Basque """

    FIN = Language.FIN
    """ Finnish """

    FRA = Language.FRA
    """ French """

    GLG = Language.GLG
    """ Galician """

    HEB = Language.HEB
    """ Hebrew """

    HIN = Language.HIN
    """ Hindi """

    HUN = Language.HUN
    """ Hungarian """

    IND = Language.IND
    """ Indonesian """

    ITA = Language.ITA
    """ Italian """

    JPN = Language.JPN
    """ Japanese """

    KOR = Language.KOR
    """ Korean """

    NLD = Language.NLD
    """ Dutch """

    NOR = Language.NOR
    """ Norwegian """

    POL = Language.POL
    """ Polish """

    POR = Language.POR
    """ Portuguese """

    RON = Language.RON
    """ Romanian """

    RUS = Language.RUS
    """ Russian """

    SLK = Language.SLK
    """ Slovak """

    SPA = Language.SPA
    """ Spanish """

    SWE = Language.SWE
    """ Swedish """

    THA = Language.THA
    """ Thai """

    TUR = Language.TUR
    """ Turkish """

    YUE = Language.YUE
    """ Yue Chinese """

    CMN_CHN = "cmn-CHN"
    """ Mandarin Chinese (China) """

    CMN_TWN = "cmn-TWN"
    """ Mandarin Chinese (Taiwan) """

    ENG_AUS = "eng-AUS"
    """ English (Australia) """

    ENG_GBR = "eng-GBR"
    """ English (GB) """

    ENG_IND = "eng-IND"
    """ English (India) """

    ENG_IRL = "eng-IRL"
    """ English (Ireland) """

    ENG_SCT = "eng-SCT"
    """ English (Scotland) """

    ENG_ZAF = "eng-ZAF"
    """ English (South Africa) """

    ENG_USA = "eng-USA"
    """ English (USA) """

    FRA_CAN = "fra-CAN"
    """ French (Canada) """

    NLD_BEL = "nld-BEL"
    """ Dutch (Belgium) """

    POR_BRA = "por-BRA"
    """ Portuguese (Brazil) """

    POR_PRT = "por-PRT"
    """ Portuguese (Portugal) """

    SPA_COL = "spa-COL"
    """ Spanish (Colombia) """

    SPA_ESP = "spa-ESP"
    """ Spanish (Spain) """

    SPA_MEX = "spa-MEX"
    """ Spanish (Mexico) """

    CODE_TO_HUMAN = {
        ARA: u"Arabic",
        CAT: u"Catalan",
        CES: u"Czech",
        CMN: u"Mandarin Chinese",
        DAN: u"Danish",
        DEU: u"German",
        ELL: u"Greek (Modern)",
        ENG: u"English",
        EUS: u"Basque",
        FIN: u"Finnish",
        FRA: u"French",
        GLG: u"Galician",
        HEB: u"Hebrew",
        HIN: u"Hindi",
        HUN: u"Hungarian",
        IND: u"Indonesian",
        ITA: u"Italian",
        JPN: u"Japanese",
        KOR: u"Korean",
        NLD: u"Dutch",
        NOR: u"Norwegian",
        POL: u"Polish",
        POR: u"Portuguese",
        RON: u"Romanian",
        RUS: u"Russian",
        SLK: u"Slovak",
        SPA: u"Spanish",
        SWE: u"Swedish",
        THA: u"Thai",
        TUR: u"Turkish",
        YUE: u"Yue Chinese",
        CMN_CHN: u"Mandarin Chinese (China)",
        CMN_TWN: u"Mandarin Chinese (Taiwan)",
        ENG_AUS: u"English (Australia)",
        ENG_GBR: u"English (GB)",
        ENG_IND: u"English (India)",
        ENG_IRL: u"English (Ireland)",
        ENG_SCT: u"English (Scotland)",
        ENG_USA: u"English (USA)",
        ENG_ZAF: u"English (South Africa)",
        FRA_CAN: u"French (Canada)",
        NLD_BEL: u"Dutch (Belgium)",
        POR_BRA: u"Portuguese (Brazil)",
        POR_PRT: u"Portuguese (Portugal)",
        SPA_COL: u"Spanish (Colombia)",
        SPA_ESP: u"Spanish (Spain)",
        SPA_MEX: u"Spanish (Mexico)",
    }

    CODE_TO_HUMAN_LIST = sorted([u"%s\t%s" % (k, v) for k, v in CODE_TO_HUMAN.items()])

    LANGUAGE_TO_VOICE_CODE = {
        ARA: "Laila",           # F, M: Maged, Tarik
        CAT: "Montserrat",      # F, M: Jordi
        CES: "Iveta",           # F, F: Zuzana
        CMN: "Tian-Tian",       # F
        DAN: "Ida",             # F, M: Magnus
        DEU: "Anna-ML",         # F-ML, F-ML: Petra-ML, M: Markus, Yannick
        ELL: "Melina",          # F, M: Nikos
        ENG: "Kate",            # F
        EUS: "Miren",           # F
        FIN: "Satu",            # F
        FRA: "Audrey-ML",       # F-ML, F: Aurelie, M: Thomas
        GLG: "Carmela",         # F
        HEB: "Carmit",          # F
        HIN: "Lekha",           # F
        HUN: "Mariska",         # F
        IND: "Damayanti",       # F
        ITA: "Alice-ML",        # F-ML, F: Federica, Paola, M: Luca
        JPN: "Kyoko",           # F, M: Otoya
        KOR: "Sora",            # F
        NLD: "Claire",          # F, M: Xander
        NOR: "Nora",            # F, M: Henrik
        POL: "Ewa",             # F, F: Zosia
        POR: "Catarina",        # F
        RON: "Ioana",           # F
        RUS: "Katya",           # F, F: Milena, M: Yuri
        SLK: "Laura",           # F
        SPA: "Monica",          # F, M: Jorge
        SWE: "Alva",            # F, M: Oskar
        THA: "Kanya",           # F
        TUR: "Yelda",           # F, M: Cem
        YUE: "Sin-Ji",          # F
        CMN_CHN: "Tian-Tian",   # F
        CMN_TWN: "Mei-Jia",     # F
        FRA_CAN: "Amelie",      # F, F: Chantal, M: Nicolas
        ENG_AUS: "Karen",       # F, M: Lee
        ENG_GBR: "Kate",        # F, F: Serena, M: Daniel, Oliver
        ENG_IND: "Veena",       # F
        ENG_IRL: "Moira",       # F
        ENG_SCT: "Fiona",       # F
        ENG_USA: "Ava",         # F, F: Allison, Samantha, Susan, Zoe, M: Tom
        ENG_ZAF: "Tessa",       # F
        NLD_BEL: "Ellen",       # F
        POR_BRA: "Luciana",     # F, M: Felipe
        POR_PRT: "Catarina",    # F, F: Joana
        SPA_COL: "Soledad",     # F, M: Carlos
        SPA_ESP: "Monica",      # F, F (Valencian): Empar
        SPA_MEX: "Angelica",    # F, F: Paulina, M: Juan
    }
    DEFAULT_LANGUAGE = ENG_GBR

    OUTPUT_AUDIO_FORMAT = ("pcm_s16le", 1, 16000)

    HAS_PYTHON_CALL = True

    # Nuance TTS API specific
    END_POINT = "NMDPTTSCmdServlet/tts"
    """ Nuance TTS API end point """

    SAMPLE_RATE = 16000
    """ Synthesize 16kHz PCM16 mono """

    URL = "https://tts.nuancemobility.net"
    """ Nuance TTS API URL """

    TAG = u"NuanceTTSWrapper"

    def __init__(self, rconf=None, logger=None):
        super(NuanceTTSWrapper, self).__init__(rconf=rconf, logger=logger)

    def _synthesize_single_python_helper(self, text, voice_code, output_file_path=None, return_audio_data=True):
        self.log(u"Importing requests...")
        import requests
        self.log(u"Importing requests... done")

        # prepare request header and contents
        request_id = str(uuid.uuid4()).replace("-", "")[0:16]
        headers = {
            u"Content-Type": u"text/plain; charset=utf-8",
            u"Accept": u"audio/x-wav;codec=pcm;bit=16;rate=%d" % self.SAMPLE_RATE
        }
        text_to_synth = text.encode("utf-8")
        url = "%s/%s?appId=%s&appKey=%s&id=%s&voice=%s" % (
            self.URL,
            self.END_POINT,
            self.rconf[RuntimeConfiguration.NUANCE_TTS_API_ID],
            self.rconf[RuntimeConfiguration.NUANCE_TTS_API_KEY],
            request_id,
            voice_code
        )

        # post request
        sleep_delay = self.rconf[RuntimeConfiguration.TTS_API_SLEEP]
        attempts = self.rconf[RuntimeConfiguration.TTS_API_RETRY_ATTEMPTS]
        self.log([u"Sleep delay:    %.3f", sleep_delay])
        self.log([u"Retry attempts: %d", attempts])
        while attempts > 0:
            self.log(u"Sleeping to throttle API usage...")
            time.sleep(sleep_delay)
            self.log(u"Sleeping to throttle API usage... done")
            self.log(u"Posting...")
            try:
                response = requests.post(url, data=text_to_synth, headers=headers)
            except Exception as exc:
                self.log_exc(u"Unexpected exception on HTTP POST. Are you offline?", exc, True, ValueError)
            self.log(u"Posting... done")
            status_code = response.status_code
            self.log([u"Status code: %d", status_code])
            if status_code == 200:
                self.log(u"Got status code 200, break")
                break
            else:
                self.log_warn(u"Got status code other than 200, retry")
                attempts -= 1

        if attempts <= 0:
            self.log_exc(u"All API requests returned status code != 200", None, True, ValueError)

        # save to file if requested
        if output_file_path is None:
            self.log(u"output_file_path is None => not saving to file")
        else:
            self.log(u"output_file_path is not None => saving to file...")
            import wave
            output_file = wave.open(output_file_path, "wb")
            output_file.setframerate(self.SAMPLE_RATE)  # sample rate
            output_file.setnchannels(1)                 # 1 channel, i.e. mono
            output_file.setsampwidth(2)                 # 16 bit/sample, i.e. 2 bytes/sample
            output_file.writeframes(response.content)
            output_file.close()
            self.log(u"output_file_path is not None => saving to file... done")

        # get length and data
        audio_sample_rate = self.SAMPLE_RATE
        number_of_frames = len(response.content) / 2
        audio_length = TimeValue(number_of_frames / audio_sample_rate)
        self.log([u"Response (bytes): %d", len(response.content)])
        self.log([u"Number of frames: %d", number_of_frames])
        self.log([u"Audio length (s): %.3f", audio_length])
        audio_format = "pcm16"
        audio_samples = numpy.fromstring(response.content, dtype=numpy.int16).astype("float64") / 32768

        # return data
        return (True, (audio_length, audio_sample_rate, audio_format, audio_samples))



#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
aeneas.cwave is a Python C extension to read WAVE mono files.

.. function:: cwave.get_audio_info(audio_file_path)

    Read the sample rate and length of the given WAVE mono file.

    The returned tuple ``(sr, length)`` contains
    the sample rate and the number of samples
    of the WAVE file.

    :param string audio_file_path: the path of the WAVE file to be read, UTF-8 encoded
    :rtype: tuple

.. function:: cwave.read_audio_data(audio_file_path, from_sample, num_samples)

    Read audio samples from the given WAVE mono file.

    The returned tuple ``(sr, data)`` contains
    the sample rate of the WAVE file,
    and the samples read as a NumPy 1D array
    of ``float64`` values in ``[-1.0, 1.0]``.

    :param string audio_file_path: the path of the WAVE file to be read, UTF-8 encoded
    :param int from_sample: index of the first sample to be read
    :param int num_samples: number of samples to be read
    :rtype: tuple
"""



#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Read audio file properties.
"""

from __future__ import absolute_import
from __future__ import print_function
import sys

from aeneas.audiofile import AudioFile
from aeneas.audiofile import AudioFileProbeError
from aeneas.audiofile import AudioFileUnsupportedFormatError
from aeneas.runtimeconfiguration import RuntimeConfiguration
from aeneas.tools.abstract_cli_program import AbstractCLIProgram
import aeneas.globalfunctions as gf


class ReadAudioCLI(AbstractCLIProgram):
    """
    Read audio file properties.
    """
    AUDIO_FILE = gf.relative_path("res/audio.mp3", __file__)

    NAME = gf.file_name_without_extension(__file__)

    HELP = {
        "description": u"Read audio file properties.",
        "synopsis": [
            (u"AUDIO_FILE", True)
        ],
        "options": [
            u"-f, --full : load samples from file, possibly converting to WAVE"
        ],
        "parameters": [
        ],
        "examples": [
            u"%s" % (AUDIO_FILE)
        ]
    }

    def perform_command(self):
        """
        Perform command and return the appropriate exit code.

        :rtype: int
        """
        if len(self.actual_arguments) < 1:
            return self.print_help()
        audio_file_path = self.actual_arguments[0]

        try:
            audiofile = AudioFile(audio_file_path, rconf=self.rconf, logger=self.logger)
            audiofile.read_properties()
            if self.has_option([u"-f", u"--full"]):
                audiofile.read_samples_from_file()
            self.print_generic(audiofile.__unicode__())
            return self.NO_ERROR_EXIT_CODE
        except OSError:
            self.print_error(u"Cannot read file '%s'" % (audio_file_path))
            self.print_error(u"Make sure the input file path is written/escaped correctly")
        except AudioFileProbeError:
            self.print_error(u"Unable to call the ffprobe executable '%s'" % (self.rconf[RuntimeConfiguration.FFPROBE_PATH]))
            self.print_error(u"Make sure the path to ffprobe is correct")
        except AudioFileUnsupportedFormatError:
            self.print_error(u"Cannot read properties of file '%s'" % (audio_file_path))
            self.print_error(u"Make sure the input file has a format supported by ffprobe")

        return self.ERROR_EXIT_CODE


def main():
    """
    Execute program.
    """
    ReadAudioCLI().run(arguments=sys.argv)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the following classes:

* :class:`~aeneas.sd.SD`, for detecting the audio head and tail of a given audio file.

.. warning:: This module is likely to be refactored in a future version

.. versionadded:: 1.2.0
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy

from aeneas.audiofilemfcc import AudioFileMFCC
from aeneas.dtw import DTWAligner
from aeneas.exacttiming import Decimal
from aeneas.exacttiming import InvalidOperation
from aeneas.exacttiming import TimeValue
from aeneas.logger import Loggable
from aeneas.runtimeconfiguration import RuntimeConfiguration
from aeneas.synthesizer import Synthesizer
import aeneas.globalfunctions as gf


class SD(Loggable):
    """
    The SD ("start detector").

    Given an audio file and a text, detects the audio head and/or tail,
    using a voice activity detector (via :class:`~aeneas.vad.VAD`) and
    performing an alignment with a partial portion of the text
    (via :class:`~aeneas.dtw.DTWAligner`).

    This implementation relies on the following heuristic:

    1. synthesize text until
       ``max_head_length`` times :data:`aeneas.sd.SD.QUERY_FACTOR`
       seconds are reached;
    2. consider only the first
       ``max_head_length`` times :data:`aeneas.sd.SD.AUDIO_FACTOR`
       seconds of the audio file;
    3. compute the best partial alignment of 1. with 2., and return
       the corresponding time value.

    (Similarly for the audio tail.)

    :param real_wave_mfcc: the audio file
    :type  real_wave_mfcc: :class:`~aeneas.audiofile.AudioFileMFCC`
    :param text_file: the text file
    :type  text_file: :class:`~aeneas.textfile.TextFile`
    :param rconf: a runtime configuration
    :type  rconf: :class:`~aeneas.runtimeconfiguration.RuntimeConfiguration`
    :param logger: the logger object
    :type  logger: :class:`~aeneas.logger.Logger`
    """

    QUERY_FACTOR = Decimal("1.0")
    """
    Multiply the max head/tail length by this factor
    to get the minimum query length to be synthesized.
    Default: ``1.0``.

    .. versionadded:: 1.5.0
    """

    AUDIO_FACTOR = Decimal("2.5")
    """
    Multiply the max head/tail length by this factor
    to get the minimum length in the audio that will be searched
    for.
    Set it to be at least ``1.0 + QUERY_FACTOR * 1.5``.
    Default: ``2.5``.

    .. versionadded:: 1.5.0
    """

    MAX_LENGTH = TimeValue("10.000")
    """
    Try detecting audio head or tail up to this many seconds.
    Default: ``10.000``.

    .. versionadded:: 1.2.0
    """

    MIN_LENGTH = TimeValue("0.000")
    """
    Try detecting audio head or tail of at least this many seconds.
    Default: ``0.000``.

    .. versionadded:: 1.2.0
    """

    TAG = u"SD"

    def __init__(self, real_wave_mfcc, text_file, rconf=None, logger=None):
        super(SD, self).__init__(rconf=rconf, logger=logger)
        self.real_wave_mfcc = real_wave_mfcc
        self.text_file = text_file

    def detect_interval(
            self,
            min_head_length=None,
            max_head_length=None,
            min_tail_length=None,
            max_tail_length=None
    ):
        """
        Detect the interval of the audio file
        containing the fragments in the text file.

        Return the audio interval as a tuple of two
        :class:`~aeneas.exacttiming.TimeValue` objects,
        representing the begin and end time, in seconds,
        with respect to the full wave duration.

        If one of the parameters is ``None``, the default value
        (``0.0`` for min, ``10.0`` for max) will be used.

        :param min_head_length: estimated minimum head length
        :type  min_head_length: :class:`~aeneas.exacttiming.TimeValue`
        :param max_head_length: estimated maximum head length
        :type  max_head_length: :class:`~aeneas.exacttiming.TimeValue`
        :param min_tail_length: estimated minimum tail length
        :type  min_tail_length: :class:`~aeneas.exacttiming.TimeValue`
        :param max_tail_length: estimated maximum tail length
        :type  max_tail_length: :class:`~aeneas.exacttiming.TimeValue`
        :rtype: (:class:`~aeneas.exacttiming.TimeValue`, :class:`~aeneas.exacttiming.TimeValue`)
        :raises: TypeError: if one of the parameters is not ``None`` or a number
        :raises: ValueError: if one of the parameters is negative
        """
        head = self.detect_head(min_head_length, max_head_length)
        tail = self.detect_tail(min_tail_length, max_tail_length)
        begin = head
        end = self.real_wave_mfcc.audio_length - tail
        self.log([u"Audio length: %.3f", self.real_wave_mfcc.audio_length])
        self.log([u"Head length:  %.3f", head])
        self.log([u"Tail length:  %.3f", tail])
        self.log([u"Begin:        %.3f", begin])
        self.log([u"End:          %.3f", end])
        if (begin >= TimeValue("0.000")) and (end > begin):
            self.log([u"Returning %.3f %.3f", begin, end])
            return (begin, end)
        self.log(u"Returning (0.000, 0.000)")
        return (TimeValue("0.000"), TimeValue("0.000"))

    def detect_head(self, min_head_length=None, max_head_length=None):
        """
        Detect the audio head, returning its duration, in seconds.

        :param min_head_length: estimated minimum head length
        :type  min_head_length: :class:`~aeneas.exacttiming.TimeValue`
        :param max_head_length: estimated maximum head length
        :type  max_head_length: :class:`~aeneas.exacttiming.TimeValue`
        :rtype: :class:`~aeneas.exacttiming.TimeValue`
        :raises: TypeError: if one of the parameters is not ``None`` or a number
        :raises: ValueError: if one of the parameters is negative
        """
        return self._detect(min_head_length, max_head_length, tail=False)

    def detect_tail(self, min_tail_length=None, max_tail_length=None):
        """
        Detect the audio tail, returning its duration, in seconds.

        :param min_tail_length: estimated minimum tail length
        :type  min_tail_length: :class:`~aeneas.exacttiming.TimeValue`
        :param max_tail_length: estimated maximum tail length
        :type  max_tail_length: :class:`~aeneas.exacttiming.TimeValue`
        :rtype: :class:`~aeneas.exacttiming.TimeValue`
        :raises: TypeError: if one of the parameters is not ``None`` or a number
        :raises: ValueError: if one of the parameters is negative
        """
        return self._detect(min_tail_length, max_tail_length, tail=True)

    def _detect(self, min_length, max_length, tail=False):
        """
        Detect the head or tail within ``min_length`` and ``max_length`` duration.

        If detecting the tail, the real wave MFCC and the query are reversed
        so that the tail detection problem reduces to a head detection problem.

        Return the duration of the head or tail, in seconds.

        :param min_length: estimated minimum length
        :type  min_length: :class:`~aeneas.exacttiming.TimeValue`
        :param max_length: estimated maximum length
        :type  max_length: :class:`~aeneas.exacttiming.TimeValue`
        :rtype: :class:`~aeneas.exacttiming.TimeValue`
        :raises: TypeError: if one of the parameters is not ``None`` or a number
        :raises: ValueError: if one of the parameters is negative
        """
        def _sanitize(value, default, name):
            if value is None:
                value = default
            try:
                value = TimeValue(value)
            except (TypeError, ValueError, InvalidOperation) as exc:
                self.log_exc(u"The value of %s is not a number" % (name), exc, True, TypeError)
            if value < 0:
                self.log_exc(u"The value of %s is negative" % (name), None, True, ValueError)
            return value

        min_length = _sanitize(min_length, self.MIN_LENGTH, "min_length")
        max_length = _sanitize(max_length, self.MAX_LENGTH, "max_length")
        mws = self.rconf.mws
        min_length_frames = int(min_length / mws)
        max_length_frames = int(max_length / mws)
        self.log([u"MFCC window shift s:     %.3f", mws])
        self.log([u"Min start length s:      %.3f", min_length])
        self.log([u"Min start length frames: %d", min_length_frames])
        self.log([u"Max start length s:      %.3f", max_length])
        self.log([u"Max start length frames: %d", max_length_frames])
        self.log([u"Tail?:                   %s", str(tail)])

        self.log(u"Synthesizing query...")
        synt_duration = max_length * self.QUERY_FACTOR
        self.log([u"Synthesizing at least %.3f seconds", synt_duration])
        tmp_handler, tmp_file_path = gf.tmp_file(suffix=u".wav", root=self.rconf[RuntimeConfiguration.TMP_PATH])
        synt = Synthesizer(rconf=self.rconf, logger=self.logger)
        anchors, total_time, synthesized_chars = synt.synthesize(
            self.text_file,
            tmp_file_path,
            quit_after=synt_duration,
            backwards=tail
        )
        self.log(u"Synthesizing query... done")

        self.log(u"Extracting MFCCs for query...")
        query_mfcc = AudioFileMFCC(tmp_file_path, rconf=self.rconf, logger=self.logger)
        self.log(u"Extracting MFCCs for query... done")

        self.log(u"Cleaning up...")
        gf.delete_file(tmp_handler, tmp_file_path)
        self.log(u"Cleaning up... done")

        search_window = max_length * self.AUDIO_FACTOR
        search_window_end = min(int(search_window / mws), self.real_wave_mfcc.all_length)
        self.log([u"Query MFCC length (frames): %d", query_mfcc.all_length])
        self.log([u"Real MFCC length (frames):  %d", self.real_wave_mfcc.all_length])
        self.log([u"Search window end (s):      %.3f", search_window])
        self.log([u"Search window end (frames): %d", search_window_end])

        if tail:
            self.log(u"Tail => reversing real_wave_mfcc and query_mfcc")
            self.real_wave_mfcc.reverse()
            query_mfcc.reverse()

        # NOTE: VAD will be run here, if not done before
        speech_intervals = self.real_wave_mfcc.intervals(speech=True, time=False)
        if len(speech_intervals) < 1:
            self.log(u"No speech intervals, hence no start found")
            if tail:
                self.real_wave_mfcc.reverse()
            return TimeValue("0.000")

        # generate a list of begin indices
        search_end = None
        candidates_begin = []
        for interval in speech_intervals:
            if (interval[0] >= min_length_frames) and (interval[0] <= max_length_frames):
                candidates_begin.append(interval[0])
            search_end = interval[1]
            if search_end >= search_window_end:
                break

        # for each begin index, compute the acm cost
        # to match the query
        # note that we take the min over the last column of the acm
        # meaning that we allow to match the entire query wave
        # against a portion of the real wave
        candidates = []
        for candidate_begin in candidates_begin:
            self.log([u"Candidate interval starting at %d == %.3f", candidate_begin, candidate_begin * mws])
            try:
                rwm = AudioFileMFCC(
                    mfcc_matrix=self.real_wave_mfcc.all_mfcc[:, candidate_begin:search_end],
                    rconf=self.rconf,
                    logger=self.logger
                )
                dtw = DTWAligner(
                    real_wave_mfcc=rwm,
                    synt_wave_mfcc=query_mfcc,
                    rconf=self.rconf,
                    logger=self.logger
                )
                acm = dtw.compute_accumulated_cost_matrix()
                last_column = acm[:, -1]
                min_value = numpy.min(last_column)
                min_index = numpy.argmin(last_column)
                self.log([u"Candidate interval: %d %d == %.3f %.3f", candidate_begin, search_end, candidate_begin * mws, search_end * mws])
                self.log([u"  Min value: %.6f", min_value])
                self.log([u"  Min index: %d == %.3f", min_index, min_index * mws])
                candidates.append((min_value, candidate_begin, min_index))
            except Exception as exc:
                self.log_exc(u"An unexpected error occurred while running _detect", exc, False, None)

        # reverse again the real wave
        if tail:
            self.log(u"Tail => reversing real_wave_mfcc again")
            self.real_wave_mfcc.reverse()

        # return
        if len(candidates) < 1:
            self.log(u"No candidates found")
            return TimeValue("0.000")
        self.log(u"Candidates:")
        for candidate in candidates:
            self.log([u"  Value: %.6f Begin Time: %.3f Min Index: %d", candidate[0], candidate[1] * mws, candidate[2]])
        best = sorted(candidates)[0][1]
        self.log([u"Best candidate: %d == %.3f", best, best * mws])
        return best * mws
